# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
tools around multipole decomposition and effective polarizabilities

"""
from __future__ import print_function
from __future__ import absolute_import

import warnings
import time
import numpy as np
import copy

from pyGDM2.multipole.main import multipole_decomposition_exact



#==============================================================================
# GLOBAL PARAMETERS
#==============================================================================
DTYPE_C = np.complex64




#==============================================================================
# EXCEPTIONS
#==============================================================================



# =============================================================================
# CONSISTENCY TESTS / HELPER
# =============================================================================
def _test_availability_generalized_polarizability(sim, which_moment, wavelength, 
                      method='lu', verbose=False, long_wavelength_approx=False):
    which_moment = which_moment.lower()
    wl = wavelength
    
    calc_gp = False
    if which_moment in ['p1', 'p0', 'p']:
        if not hasattr(sim, 'K_P_E'):
            calc_gp = True
        elif wl not in sim.K_P_E.keys():
            calc_gp = True
            
    if which_moment in ['pt', 'p']:
        if not hasattr(sim, 'K_T_E'):
            calc_gp = True
        elif wl not in sim.K_T_E.keys():
            calc_gp = True
            
    if which_moment == 'm':
        if not hasattr(sim, 'K_M_E'):
            calc_gp = True
        elif wl not in sim.K_M_E.keys():
            calc_gp = True
            
    if which_moment in ['qe1', 'qe0', 'qe']:
        if not hasattr(sim, 'K_QE_E'):
            calc_gp = True
        elif wl not in sim.K_QE_E.keys():
            calc_gp = True
            
    if which_moment in ['qet', 'qe']:
        if not hasattr(sim, 'K_QT_E'):
            calc_gp = True
        elif wl not in sim.K_QT_E.keys():
            calc_gp = True
            
    if which_moment == 'qm':
        if not hasattr(sim, 'K_QM_E'):
            calc_gp = True
        elif wl not in sim.K_QM_E.keys():
            calc_gp = True
    
    if calc_gp:
        if verbose:
            warnings.warn('generalized polarizabilities not available. evaluating...')
        generalized_polarizability(sim, wavelength=wl, method=method, 
                                   long_wavelength_approx=long_wavelength_approx)



# =============================================================================
# MODULE FUNCTIONS
# =============================================================================
def generalized_polarizability(sim, field_index=None, wavelength=None, 
                               method='lu', epsilon=0.01, r0=None,
                               which_moments=['p1', 'pt', 'qe', 'qt', 'm', 'qm'],
                               long_wavelength_approx=False, verbose=1):
    """generalized electric and magnetic polarizabilities
    
    ** ------- FUNCTION STILL UNDER TESTING ------- **
    
    Returns the generalized polarizability tensors that can be used with arbitrary,
    inhomogeneous illumination fields to calculate the effective 
    electric and magnetic multipole moments, induced in the nanostructure.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
    
    wavelength : float, default: None
        Optional wavelength (alternative to `field_index`) at which to 
        calculate susceptibility matrix (in nm). 
        Either `field_index` or `wavelength` must be given.
    
    method : string, default: "scipyinv"
        inversion method. One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
         - "scipyinv" scipy default inversion (`scipy.linalg.inv`)
         - "numpyinv" numpy inversion (`np.linalg.inv`, if numpy compiled accordingly: LAPACK's `dgesv`)
         - "cupy" uses CUDA GPU via `cupy`
         - "cuda" (equivalent to "cupy")
         - "lu" LU-decomposition (`scipy.linalg.lu_factor`) - inefficient for `decay_rate`!
    
    epsilon : float, default: 0.01
        additional step on r0 (in nm) to avoid numerical divergence of the Bessel terms
        
    r0 : array, default: None
        [x,y,z] position of mulipole decomposition development. 
        If `None`, use structure's center of gravity
    
    which_moments : list of str, default: ['p1', 'pt', 'qe1', 'qt', 'm', 'qm']
        which generalized polarizability tensors to calculate and return. supported:
            - 'p1': electric dipole - only first order (rank 2)
            - 'pt': toroidal dipole (rank 3)
            - 'qe1': electric quadrupole - only first order (rank 4)
            - 'qt': toroidal electric quadrupole (rank 4)
            - 'm': magnetic dipole (rank 2)
            - 'qm': magnetic quadrupole (rank 3)
            
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation
        
    verbose : bool default=True
        print runtime info
    
    
    Returns
    -------
    
    by default 6 lists of N tensors, with N the number of discretization cells (see kwarg `which_moments`):
        
    K_P_E, K_T_E, K_QE_E, K_QT_E, K_M_E, K_QM_E
    
    Notes
    -----
    For details , see: 
        
    *PAPER SUBMITTED*
    
    For details about the underlying exact multipole decomposition, see: 
        
    Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic 
    multipole expansion beyond the long-wavelength approximation.*
    Optics Communications 407, 17–21 (2018)
    
    """
    # =============================================================================
    #     Exception handling
    # =============================================================================
    if field_index is None and wavelength is None:
        raise Exception("Either `field_index` or `wavelength` must be given!")
        
    if field_index is not None and wavelength is not None:
        warnings.warn("`field_index` AND `wavelength` are given! Ignoring `wavelength`.")
        
    
    
    # =============================================================================
    # preparation
    # =============================================================================
    from pyGDM2 import core
    
    which_moments = [wm.lower() for wm in which_moments]
    
    if field_index is not None:
        from pyGDM2 import tools
        wavelength = tools.get_field_indices(sim)[field_index]['wavelength']
    
    if 'p1' in which_moments and not hasattr(sim, 'K_P_E'):
        sim.K_P_E = dict()
    if 'pt' in which_moments and not hasattr(sim, 'K_T_E'):
        sim.K_T_E = dict()
    if 'qe' in which_moments and not hasattr(sim, 'K_QE_E'):
        sim.K_QE_E = dict()
    if 'qt' in which_moments and not hasattr(sim, 'K_QT_E'):
        sim.K_QT_E = dict()
    if 'm' in which_moments and not hasattr(sim, 'K_M_E'):
        sim.K_M_E = dict()
    if 'qm' in which_moments and not hasattr(sim, 'K_QM_E'):
        sim.K_QM_E = dict()
    
    
    
    if verbose:
        t0 = time.time()
        print("wl={}nm. calc. K:".format(np.round(wavelength, 1)), end='')
    
    
    ## structure
    geo = sim.struct.geometry
    alpha = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    if r0 is None:
        r0 = np.average(geo, axis=0)
    sim.r0 = r0
    if np.abs(np.linalg.norm(geo - r0, axis=1)).min() > epsilon:
        epsilon = 0
    Dr = geo - r0  #   = r-r0
    norm_r = np.linalg.norm(Dr + epsilon, axis=1)  # epsilon: avoid divergence of 1/kr at r=0
    norm_r2 = norm_r**2
    
    ## illumination properties
    sim.struct.setWavelength(wavelength)
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, r0[None,:])[0]  # assume structure is fully in one environment
    n_env = eps_env**0.5
    k0 = 2.0*np.pi / wavelength
    k = k0*n_env
    kr = k * norm_r
    
    ## bessel functions and pre-factors
    if not long_wavelength_approx:
        from scipy.special import spherical_jn as sph_jn
        j0kr = sph_jn(0, kr)
        j1kr = sph_jn(1, kr) / kr
        j2kr = sph_jn(2, kr) / (kr**2)
        j3kr = sph_jn(3, kr) / (kr**3)
        f_pt = 1/2; f_ptA=3; f_ptB=-1
        f_qe = 3; fqe2=2; fqe2A = 5; fqe2B = -1; fqe2C = -1
        f_m = 3/2; f_qm = 15
    else:
        j0kr = j1kr = j2kr = j3kr = np.ones_like(kr)
        f_pt = 1/10; f_ptA=1; f_ptB=-2
        f_qe = 1; fqe2=1/14; fqe2A = 4; fqe2B = -5; fqe2C = 2
        f_m = 1/2; f_qm = 1
    
    
    ## ----- gen. propagator
    K = core.get_general_propagator(sim, wavelength, method=method)
    if method.lower() in ['cupy', 'cuda']:
        K = K.get()
    if method.lower() == 'lu':
        import scipy.linalg as la
        K = la.lu_solve(K, np.identity(K[0].shape[0], dtype=K[0].dtype))
        
    if verbose:
        t1 = time.time()
        print(" {:.1f}ms.  electric... ".format((t1-t0)*1000), end='')
    K2 = K.reshape(len(K)//3, 3, -1, 3).swapaxes(1,2).reshape(-1, 3, 3).reshape(len(K)//3,len(K)//3, 3, 3)
    K2a = np.matmul(alpha[:,None,...], K2)
    Ka = np.concatenate(np.concatenate(K2a, axis=1), axis=1)


# =============================================================================
#     electric-electric
# =============================================================================
    ## ----- electric-electric generalized polarizability - dipole
    if 'p1' in which_moments:
        K_P_E = np.sum(K2a * j0kr[:,None,None,None], axis=0)
        sim.K_P_E[wavelength] = K_P_E       # store in simulation object


    ## ----- higher order electric moments
    if 'pt' in which_moments or 'qe' in which_moments or 'qt' in which_moments:
        K_re = np.zeros_like(Ka)
        for i_dp in range(len(Ka)):
            K_re.T[i_dp] = (Dr * Ka.T[i_dp].reshape((len(Ka)//3,3))).reshape((len(Ka),))
        K2_re = K_re.reshape(len(Ka)//3, 3, -1, 3).swapaxes(1,2).reshape(-1, 3, 3).reshape(len(Ka)//3,len(Ka)//3, 3, 3)
            
        
    ## ----- toroidal dipole generalized polarizability 
    if 'pt' in which_moments:
        Qt1 = np.zeros((len(Ka)//3,3,3,3), dtype=np.complex64)
        Qt2 = np.zeros((len(Ka)//3,3,3), dtype=np.complex64)
        for i_a in range(3):
            for i_b in range(3):
                Qt1[:, i_a, i_b] = np.sum(
                    (
                    K2_re[...,i_b,:] * Dr[:,i_a,None,None]
                      ) * j2kr[:, None, None], axis=0)
            Qt2[:, i_a] = np.sum(
                (K2a[...,i_a,:]) * norm_r2[:,None,None]* j2kr[:, None, None], axis=0)
                
        K_T1_E = k**2 * f_pt * f_ptA * Qt1 
        K_T2_E = k**2 * f_pt * f_ptB * Qt2
    
        K_T_E = K_T1_E + K_T2_E[...,None,:]/3
        sim.K_T_E[wavelength] = K_T_E       # store in simulation object


    ## ----- electric quadrupole generalized polarizability 
    krondelta = np.identity(3)
    if 'qe' in which_moments:
        Qe11 = np.zeros((len(Ka)//3,3,3,3), dtype=np.complex64)
        Qe12 = np.zeros((len(Ka)//3,3,3,3,3), dtype=np.complex64)

        for i_a in range(3):
            for i_b in range(3):
                Qe11[:, i_a, i_b] = np.sum(
                    (
                    3*(Dr[:,i_b,None,None] * K2a[...,i_a,:] +
                       Dr[:,i_a,None,None] * K2a[...,i_b,:])
                      ) * j1kr[:, None,None], axis=0)
                
                for i_c in range(3):  # keep separate scalar product elements in last index
                    Qe12[:, i_a, i_b, i_c] = np.sum(
                        (
                        - 2*K2_re[...,i_c,:] * krondelta[i_a,i_b]   # diagonal term
                          ) * j1kr[:, None,None], axis=0)
        
        K_QE_E = f_qe * (Qe12 + Qe11[...,None,:]/3)
        sim.K_QE_E[wavelength] = K_QE_E
    
    
    ## ----- electric toroidal quadrupole generalized polarizability 
    if 'qt' in which_moments:
        Qet1 = np.zeros((len(Ka)//3,3,3,3,3), dtype=np.complex64)
        Qet2 = np.zeros((len(Ka)//3,3,3,3), dtype=np.complex64)
        Qet3 = np.zeros((len(Ka)//3,3,3,3,3), dtype=np.complex64)

        for i_a in range(3):
            for i_b in range(3):
                Qet2[:, i_a, i_b] = np.sum(
                    (
                    fqe2B * (Dr[:,i_a,None,None] * K2a[...,i_b,:] + Dr[:,i_b,None,None] * K2a[...,i_a,:])
                      ) * norm_r2[:,None,None] * j3kr[:, None,None], axis=0)
                
                for i_c in range(3):  # keep separate scalar product elements in last index
                    Qet1[:, i_a, i_b, i_c] = np.sum(
                        (
                         fqe2A * K2_re[...,i_c,:] * Dr[:,i_a,None,None] * Dr[:,i_b,None,None]
                          ) * j3kr[:, None,None], axis=0)
                    Qet3[:, i_a, i_b, i_c] = np.sum(
                        (
                         fqe2C * K2_re[...,i_c,:] * krondelta[i_a,i_b]  # diagonal term
                          ) * norm_r2[:,None,None] * j3kr[:, None,None], axis=0)
        K_QT_E = f_qe * fqe2 * k**2 * (Qet1 + Qet2[...,None,:]/3 + Qet3)
        sim.K_QT_E[wavelength] = K_QT_E



# =============================================================================
#     electric-magnetic
# =============================================================================
    if 'm' in which_moments or 'qm' in which_moments:
        if verbose:
            t2 = time.time()
            print("{:.1f}ms.  magnetic... ".format((t2-t1)*1000), end='')
        K_he = np.zeros_like(Ka)
        for i_dp in range(len(Ka)):
            K_he.T[i_dp] = np.cross(Dr, Ka.T[i_dp].reshape((len(Ka)//3,3))).reshape((len(Ka),))
        K2_he = K_he.reshape(len(Ka)//3, 3, -1, 3).swapaxes(1,2).reshape(-1, 3, 3).reshape(len(Ka)//3,len(Ka)//3, 3, 3)
    
    
    ## ----- electric-magnetic generalized polarizability         
    if 'm' in which_moments:
        K_M_E = np.sum(K2_he * j1kr[:, None, None, None], axis=0)
        K_M_E = -1j*k * f_m * K_M_E
        sim.K_M_E[wavelength] = K_M_E       # store in simulation object


    ## ----- electric-magnetic quadrupole generalized polarizability
    if 'qm' in which_moments:
        Qm = np.zeros((len(Ka)//3,3,3,3), dtype=np.complex64)
        for i_a in range(3):
            for i_b in range(3):
                Qm[:, i_a, i_b] = np.sum(
                    (
                    Dr[:,i_a,None,None] * K2_he[...,i_b,:]  + 
                    Dr[:,i_b,None,None] * K2_he[...,i_a,:] 
                      ) * j2kr[:, None, None], axis=0)
                
        K_QM_E = -1j*k * f_qm * Qm 
        sim.K_QM_E[wavelength] = K_QM_E       # store in simulation object
    
    
    if verbose: 
        print("{:.1f}ms. Done.".format((time.time()-t2)*1000))
    
# =============================================================================
#     return results
# =============================================================================
    return_list = []
    for _m in which_moments:
        if _m.lower() == "p1":
            return_list.append(K_P_E)
        if _m.lower() == "pt":
            return_list.append(K_T_E)
        if _m.lower() in ["qe"]:
            return_list.append(K_QE_E)
        if _m.lower() in ["qt"]:
            return_list.append(K_QT_E)
        if _m.lower() == "m":
            return_list.append(K_M_E)
        if _m.lower() in ["qm"]:
            return_list.append(K_QM_E)
    
    return return_list







def eval_generalized_polarizability_p(sim, field_index, which_order='p', method='lu', 
                                      long_wavelength_approx=False):
    """get electric dipole moment via generalized polarizability
    
    will evaluate generalized pola. and store it in simulation, in case it 
    is not yet calculated.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
        
    which_order : str, optional
        The default is 'p'. One of:
            - 'p': full dipole moment including toroidal contribution
            - 'p1': only first order dipole
            - 'pt': only toroidal dipole moment
    
    method : string, default: "lu"
        inversion method if generalized pola. not yet calculated. 
        One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation


    Returns
    -------
    electric dipole moment (complex vector)

    """
    from pyGDM2 import tools
    
    kws_wl = tools.get_field_indices(sim)[field_index]
    wl = kws_wl['wavelength']
    
    for which_moment in ['p1', 'pt']:
        _test_availability_generalized_polarizability(sim, which_moment, 
              wl, method=method, long_wavelength_approx=long_wavelength_approx)
    
        
    K_P_E = sim.K_P_E[wl]
    K_T_E = sim.K_T_E[wl]
    
    ## illuminatin field for 'field_index'
    env_dict = sim.dyads.getConfigDictG(wl, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(sim.struct.geometry, env_dict, **kws_wl)
    
    if which_order.lower() in ['p', 'p0', 'p1']:  # 'p0' equiv. to 'p1'
        p1 = np.sum(np.matmul(K_P_E, E0[...,None]), axis=(0,2))
        if which_order.lower() in ['p0', 'p1']:
            p = p1
        
    if which_order.lower() in ['p', 'pt']:
        pt = np.array([np.sum(np.matmul(K_T_E[...,i,:,:], E0[:,:,None]), axis=(0,2)) for i in range(3)])
        pt = np.sum(pt, axis=1)
        if which_order.lower() == 'p':
            p = p1 + pt
        else:
            p = pt
    return p


def eval_generalized_polarizability_qe(sim, field_index, which_order='qe', method='lu', 
                                      long_wavelength_approx=False):
    """get electric quadrupole moment via generalized polarizability
    
    will evaluate generalized pola. and store it in simulation, in case it 
    is not yet calculated.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
        
    which_order : str, optional
        The default is 'qe'. One of:
            - 'qe': full electric quadrupole moment (including toroidal contribution)
            - 'qe1': only first order contribution to quadrupole
            - 'qet': only toroidal quadrupole moment
            
    method : string, default: "lu"
        inversion method if generalized pola. not yet calculated. 
        One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation

        
    Returns
    -------
    electric quadrupole moment (complex tensor, 3x3)

    """
    from pyGDM2 import tools
    
    kws_wl = tools.get_field_indices(sim)[field_index]
    wl = kws_wl['wavelength']
    
    for which_moment in ['qe', 'qt']:
        _test_availability_generalized_polarizability(sim, which_moment, 
              wl, method=method, long_wavelength_approx=long_wavelength_approx)
    
        
    K_QE_E = sim.K_QE_E[wl]
    K_QT_E = sim.K_QT_E[wl]
    
    ## illumination field for 'field_index'
    env_dict = sim.dyads.getConfigDictG(wl, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(sim.struct.geometry, env_dict, **kws_wl)
    
    
    if which_order.lower() in ['q', 'qe', 'qe0', 'qe1']:  # 'qe0' equiv. to 'qe1'
        qe1 = np.array([
            [np.sum(np.matmul(K_QE_E[...,i,j,:,:], E0[:,:,None]), axis=(0,2)) 
                                         for j in range(3)] for i in range(3)])
        qe1 = np.sum(qe1, axis=-1)
        if which_order.lower() in ['qe0', 'qe1']:
            qe = qe1
        
    if which_order.lower() in ['q', 'qe', 'qet', 'qt']:
        qet = np.array([
            [np.sum(np.matmul(K_QT_E[...,i,j,:,:], E0[:,:,None]), axis=(0,2)) 
                                         for j in range(3)] for i in range(3)])
        qet = np.sum(qet, axis=-1)
        if which_order.lower() in ['q', 'qe']:  # 'q' equiv. to 'qe'
            qe = qe1 + qet
        else:
            qe = qet
    return qe


def eval_generalized_polarizability_m(sim, field_index, method='lu', 
                                      long_wavelength_approx=False):
    """get magnetic dipole moment via generalized polarizability
    
    will evaluate generalized pola. and store it in simulation, in case it 
    is not yet calculated.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
            
    method : string, default: "lu"
        inversion method if generalized pola. not yet calculated. 
        One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation

        
    Returns
    -------
    magnetic quadrupole moment (complex vector)

    """
    from pyGDM2 import tools
    
    kws_wl = tools.get_field_indices(sim)[field_index]
    wl = kws_wl['wavelength']
    
    for which_moment in ['m']:
        _test_availability_generalized_polarizability(sim, which_moment, 
              wl, method=method, long_wavelength_approx=long_wavelength_approx)     
    
    K_M_E = sim.K_M_E[wl]
    
    ## illuminatin field for 'field_index'
    env_dict = sim.dyads.getConfigDictG(wl, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(sim.struct.geometry, env_dict, **kws_wl)
    
    m = np.sum(np.matmul(K_M_E, E0[...,None]), axis=(0,2))
    
    return m


def eval_generalized_polarizability_qm(sim, field_index, method='lu', 
                                      long_wavelength_approx=False):
    """get magnetic quadrupole moment via generalized polarizability
    
    will evaluate generalized pola. and store it in simulation, in case it 
    is not yet calculated.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
            
    method : string, default: "lu"
        inversion method if generalized pola. not yet calculated. 
        One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation

        
    Returns
    -------
    magnetic quadrupole tensor (complex tensor, 3x3)

    """
    from pyGDM2 import tools
    
    kws_wl = tools.get_field_indices(sim)[field_index]
    wl = kws_wl['wavelength']
    
    for which_moment in ['qm']:
        _test_availability_generalized_polarizability(sim, which_moment, 
              wl, method=method, long_wavelength_approx=long_wavelength_approx)
    
    K_QM_E = sim.K_QM_E[wl]
    
    ## illuminatin field for 'field_index'
    env_dict = sim.dyads.getConfigDictG(wl, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(sim.struct.geometry, env_dict, **kws_wl)
    
    qm = np.array([np.sum(np.matmul(K_QM_E[...,i,:,:], E0[:,:,None]), axis=(0,2)) for i in range(3)])
    
    return qm


def extract_effective_polarizability(sim, method='lu',
                                     which_moments=['p1','m'], long_wavelength_approx=True,
                                     illumination_mode='dipole', npoints=10, r_sphere=5000,
                                     store_simulation_object=False):
    """Extract effective electric and magnetic dipole polarizability for structure
    
    solve inverse problem of adjusting polarizability for different illuminations
    via pseudoinverse
    
    *doc to be completed*
    
    store_simulation_object : Bool, default: False
        optionally store full sim in output dictionnary. 
        *Caution:* This contains numba compiled objects which are not 
        necessarily reloadable on other operating systems than the 
        one used for the simulation

    """
    from pyGDM2 import fields
    from pyGDM2 import tools
    from scipy import linalg
    
    def sample_spherical(npoints, ndim=3):
        """random pos. on sphere (R=1). from: https://stackoverflow.com/questions/33976911"""
        vec = np.random.randn(ndim, npoints) - 0.5
        vec /= np.linalg.norm(vec, axis=0)
        return vec

    ## dipoles of random position and random orientation
    wavelengths = sim.efield.wavelengths
    
    r0 = np.average(sim.struct.geometry, axis=0)
    sim.r0 = r0
    

    if illumination_mode.lower() == 'dipole':
        rnd_pos = sample_spherical(npoints, ndim=3).T * r_sphere
        field_kwargs = [
            dict(x0=rnd_pos[i,0], y0=rnd_pos[i,1], z0=rnd_pos[i,2], 
                 mx=np.random.random()*r_sphere,
                 my=np.random.random()*r_sphere,
                 mz=np.random.random()*r_sphere)
            for i in range(npoints)
            ]
        field_generator = fields.dipole_electric
    else:
    ## -- alternative: plane wave with different polarizations / incident angles
        field_kwargs = [
            dict(inc_angle=0, inc_plane='xz', E_s=0.0, E_p=1.0),
            dict(inc_angle=0, inc_plane='xz', E_s=1.0, E_p=0.0),
            dict(inc_angle=90, inc_plane='xz', E_s=1.0, E_p=0.0),
            dict(inc_angle=90, inc_plane='xz', E_s=0.0, E_p=1.0),
            dict(inc_angle=90, inc_plane='yz', E_s=1.0, E_p=0.0),
            dict(inc_angle=90, inc_plane='yz', E_s=0.0, E_p=1.0),
            ]
        field_generator = fields.plane_wave
    
    n_illum = len(field_kwargs)
    efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=copy.deepcopy(field_kwargs))
    _sim = sim.copy()
    _sim.efield = efield
    _sim.scatter(method=method)
    
    ahh = []
    aee = []
    for i_wl, wl in enumerate(wavelengths):
        p_eval = np.zeros((3, n_illum), dtype=np.complex64)
        m_eval = np.zeros((3, n_illum), dtype=np.complex64)
        E0_eval = np.zeros((3, n_illum), dtype=np.complex64)
        H0_eval = np.zeros((3, n_illum), dtype=np.complex64)
        for idx, kw  in enumerate(field_kwargs):
            ## illuminatin field for 'field_index'
            env_dict = _sim.dyads.getConfigDictG(wl, _sim.struct, _sim.efield)
            E0 = efield.field_generator([_sim.r0], env_dict, wavelength=wl, **kw)[0]
            # inc. field array shape: (3, n_fields)
            E0_eval[:, idx] = E0
            
            H0 = efield.field_generator([_sim.r0], env_dict, wavelength=wl, returnField='H', **kw)[0]
            H0_eval[:, idx] = H0
            
            kw_wl = kw.copy()
            kw_wl['wavelength'] = wl
            field_index = tools.get_closest_field_index(_sim, kw_wl)
            p, m = multipole_decomposition_exact(_sim, field_index, which_moments=which_moments, 
                                                 long_wavelength_approx=long_wavelength_approx)
            p_eval[...,idx] = p
            m_eval[...,idx] = m
        
        ## pseudo-inverse of all illuminations
        AEinv = linalg.pinv(np.conj(E0_eval).T)
        AHinv = linalg.pinv(np.conj(H0_eval).T)
        
        ## optimum alphas to obtain dipole moments for each illumination
        alpha_pinv = np.conj(np.dot(AEinv, np.conj(p_eval).T)).T
        alpha_minv = np.conj(np.dot(AHinv, np.conj(m_eval).T)).T
        
        aee.append(alpha_pinv.reshape([3,3]))
        ahh.append(alpha_minv.reshape([3,3]))
    
    enclosing_radius = tools.get_enclosing_sphere_radius(sim)
    
    dict_pola_pseudo = dict(
                r0=sim.r0, r0_MD=sim.r0, r0_ED=sim.r0, 
                full_geometry=sim.struct.geometry, 
                alpha_pE=aee, 
                alpha_mH=ahh, 
                wavelengths=sim.efield.wavelengths, 
                enclosing_radius=enclosing_radius,
                k0_spectrum=2 * np.pi / np.array(wavelengths).copy(),
                )
    
    # optionally add also the full sim.
    if store_simulation_object:
        dict_pola_pseudo['sim'] = sim
    
    return dict_pola_pseudo




# encoding: utf-8
from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import core
from pyGDM2 import visu
from pyGDM2 import tools
from pyGDM2 import linear


## --- simulation initialization ---
## structure: dielectric sphere of 60nm radius in vacuum
step = 20
geometry = structures.sphere(step, R=3, mesh='cube')
material = materials.dummy(3.0)
norm = structures.get_normalization(mesh='cube')
n1 = n2 = 1.0

struct = structures.struct(step, geometry, material, n1,n2, norm)

## incident field
field_generator = fields.planewave
wavelengths = [400, 600, 800, 1000, 1200]
kwargs = dict(theta=0.0, kSign=-1)
efield = fields.efield(field_generator, 
               wavelengths=wavelengths, kwargs=kwargs)

## create simulation object
sim = core.simulation(struct, efield)




#%%
## you can make core.scatter call a user-defined function upon every evalutaed
def userdefined_callback_func(callback_dict):
    print("dictionary passed to callback contains:")
    print(callback_dict.keys())
    print("last calculated wavelngth: {}nm".format(callback_dict['wavelength']))
    
    print("calculate nearfield above struct for latest wavelength:")
    sim_tmp = callback_dict['sim']   # unfinished simulation, containing only the so far calculated fields
    r_probe = tools.generate_NF_map(-200, 200, 21, -200, 200, 21, Z0=200)
    fidx = tools.get_closest_field_index(sim_tmp, dict(wavelength=callback_dict['wavelength']))
    Es, Et, Bs, Bt = linear.nearfield(sim, fidx, r_probe)
    visu.vectorfield_color(Et, tit='wl={}nm'.format(callback_dict['wavelength']))
    
    
    if callback_dict['wavelength']<900:
        print ("wavelength<900nm. continue...\n")
        return True   # return True means continue
    ## abort simulation if last wavelength > 900nm
    else:
        print ("\n")
        print ("wl>900nm. abort simulation to demonstrate use of stop condition.")
        return False  # if return False here, core.scatter will stop
    


## --- run the simulation ---
core.scatter(sim, callback=userdefined_callback_func, verbose=False)





# encoding: utf-8
"""
Example script, comparing pyGDM with Mie theory for a gold sphere

"""
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import tools
from pyGDM2 import linear
from pyGDM2 import visu




## --- load pre-calculated Mie-data
wl_mie, qext_mie, qsca_mie = np.loadtxt("scat_mie_Au_D50nm.txt").T
qabs_mie = qext_mie - qsca_mie


#==============================================================================
# pyGDM setup
#==============================================================================
## --- Setup incident field
field_generator = fields.planewave
## log-interval spectrum (denser at low lambda):
wavelengths = np.exp(np.linspace(np.log(400), np.log(700), 30))
kwargs = dict(theta = [0.0])
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## --- Setup geometry (sphere D=50nm in vacuum)
scale_factor = 1.38     # this value was used in doc.-paper example
step = 6.25/scale_factor
radius = 4.*scale_factor
geometry = structures.sphere(step, R=radius, mesh='hex', ORIENTATION=2)

material = materials.gold()
n1, n2 = 1.0, 1.0     # vacuum env.

struct = structures.struct(step, geometry, material, n1,n2, 
                                   structures.get_normalization('hex'))


sim = core.simulation(struct, efield)

visu.structure(sim)
print('(hex) ----- N_dipoles =', len(sim.struct.geometry), end='')

#%%
#==============================================================================
# run the simulations
#==============================================================================
# E = core.scatter(sim, method='lu', verbose=True)
E = core.scatter(sim, method='cupy', matrix_setup='numba', verbose=True)

## extinction spectrum
field_kwargs = tools.get_possible_field_params_spectra(sim)[0]
wl, spec = tools.calculate_spectrum(sim, field_kwargs, linear.extinct)
a_ext, a_sca, a_abs = spec.T
a_geo = tools.get_geometric_cross_section(sim)



#%%
#==============================================================================
# plot
#==============================================================================
plt.figure(figsize=(5,4), dpi=100)
plt.title("gold sphere, D=50nm")

## --- Mie
plt.plot(wl_mie, qext_mie, 'b--', dashes=[2,1],label='ext.')
plt.plot(wl_mie, qabs_mie, 'g--', dashes=[2,1],label='abs.')
plt.plot(wl_mie, qsca_mie, 'r--', dashes=[2,1],label='scat.')

## --- pyGDM
plt.scatter(wl, a_ext/a_geo, marker='x', linewidth=1.5,s=25, color='b', label='')
plt.scatter(wl, a_abs/a_geo, marker='x', linewidth=1.5,s=25, color='g', label='')
plt.scatter(wl, a_sca/a_geo, marker='x', linewidth=1.5,s=25, color='r', label='')

plt.xlim(400,700)

## --- for legend only
plt.plot([0], [0], 'k--', dashes=[2,1], label='Mie')
plt.scatter([0], [0], marker='x', linewidth=1.5, color='k', label='pyGDM')
## -- legend
plt.legend(loc='best', fontsize=12)


plt.xlabel("wavelength (nm)")
plt.ylabel("ext. / scat. / abs. efficiency")
# plt.xlim( [wl.min(), wl.max()] )
plt.ylim( [0, 1.7] )



plt.tight_layout()
plt.show()



# encoding: utf-8
#
# Copyright (C) 2017-2024, P. R. Wiecha
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
Collection of incident fields

authors: P. Wiecha, C. Majorel
"""

from __future__ import print_function
from __future__ import absolute_import

import warnings
import types
import cmath

import numpy as np
import numba


# ==============================================================================
# globals
# ==============================================================================
DTYPE_C = np.complex64


# ==============================================================================
# field generator functions
# ==============================================================================
def nullfield(pos, env_dict, wavelength, returnField="E", **kwargs):
    """Zero-Field

    all additional kwargs are ignored

    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]

    env_dict : dict
        placeholder for environment description. will be ignored in `nullfield`

    wavelength : float
        Wavelength in nm

    returnField : str
        placeholder. will be ignored in `nullfield`

    Returns:
    ----------
      Complex zero 3-vector at each coordinate in 'pos'

    """
    return np.zeros((len(pos), 3), dtype=DTYPE_C)


##----------------------------------------------------------------------
##                      INCIDENT FIELDS
##----------------------------------------------------------------------
@numba.njit(cache=True)
def _three_layer_pw(
    wavelength, theta_inc, polar, z_d, spacing, n1, n2, n3, x, y, z, E0=1.0
):
    """oblique incident planewave, only linear polarization

    Oblique incidence (from bottom to top) through n1/n2/n3 layer interfaces.
    May be used to simulate evanescent fields in the total internal
    reflection configuration. Linear polarization.
    Amplitude = 1 for both, E and B.

    Original code by Ch. Girard, python implementation by C. Majorel

    Parameters
    ----------
    wavelength : float
        Wavelength in nm

    theta_inc : float, default: 0
        incident angle in the XZ plane with respect to e_z, in degrees.
         - 0deg = along Z (from neg to pos Z)
         - 90deg = along X  (from pos to neg X)
         - 180deg = along Z  (from pos to neg Z)
         - 270deg = along X  (from neg to pos X)

    polar : str, default: 's'
        incident linear polarization. Either 's' or 'p'.
        At 0 degree incident angle, 's' is polarized along x, 'p' along y.

    z_d : float
        bottom interface position (along z) between media 1 and 2

    spacing : float
        spacing between bottom interface (between media 1 and 2) and
        top interface position (between media 2 and 3)

    n1, n2, n3 : complex
        (complex) refractive index of each media

    x, y, z : float
        x/y/z coordinates for computation of the fields

    E0 : float
        amplitude of E (and B) field

    Returns
    -------
      E0, B0:       Complex E-B-Fields at each dipole position as
                    6 lists of the (complex) components:
                    Ex, Ey, Ez, Bx, By, Bz
    """
    if E0 == 0.0:
        return 0.0j, 0.0j, 0.0j, 0.0j, 0.0j, 0.0j

    z_u = z_d + spacing  # z position of upper interface
    r = np.array([x, y, z]).astype(np.complex64)  # eval. position
    theta_r = theta_inc * np.pi / 180.0  # inc. angle in rad
    k0 = 2 * np.pi / wavelength  # wavevector in vacuum

    ## -- permittivities
    eps1 = n1**2
    eps2 = n2**2
    eps3 = n3**2

    if theta_r > np.pi / 2.0 and theta_r < 3 * np.pi / 2.0:
        ## -- Different wavevectors
        k1 = np.array(
            [
                -n3 * k0 * cmath.sin(theta_r),
                0.0,
                -k0 * cmath.sqrt(eps1 - eps3 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- transmitted wavevector in medium 1 (bottom layer)
        k2 = np.array(
            [
                -n3 * k0 * cmath.sin(theta_r),
                0.0,
                -k0 * cmath.sqrt(eps2 - eps3 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- transmitted wavevector in medium 2 (middle layer)
        k2p = np.array(
            [
                -n3 * k0 * cmath.sin(theta_r),
                0.0,
                k0 * cmath.sqrt(eps2 - eps3 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- reflected wavevector in medium 2 (middle layer)
        k3 = np.array(
            [-n3 * k0 * cmath.sin(theta_r), 0.0, n3 * k0 * cmath.cos(theta_r)]
        ).astype(
            np.complex64
        )  ## -- incident wavevector in medium 3 (top layer)
        k3p = np.array(
            [-n3 * k0 * cmath.sin(theta_r), 0.0, -n3 * k0 * cmath.cos(theta_r)]
        ).astype(
            np.complex64
        )  ## -- reflected wavevector in medium 3 (top layer)

        ## -- Phase terms
        c1p = cmath.exp(1.0j * k1[2] * z_d)
        c1m = cmath.exp(-1.0j * k1[2] * z_d)
        c2p = cmath.exp(1.0j * k2[2] * z_d)
        c2m = cmath.exp(-1.0j * k2[2] * z_d)
        c2pp = cmath.exp(1.0j * k2[2] * z_u)
        c2pm = cmath.exp(-1.0j * k2[2] * z_u)
        c3pp = cmath.exp(1.0j * k3[2] * z_u)
        c3pm = cmath.exp(-1.0j * k3[2] * z_u)

        ## -- z - components of the wavevector/eps for magnetic modulus
        k1gz = k1[2] / eps1
        k2gz = k2[2] / eps2
        k3gz = k3[2] / eps3

        ### --- modulus electric field in s-polarized mode
        delta = (
            c3pm
            * c1p
            * (
                c2m
                * c2pp
                * (k2[2] ** 2 + k1[2] * k3[2] + k3[2] * k2[2] + k1[2] * k2[2])
                + c2p
                * c2pm
                * (-k2[2] ** 2 - k1[2] * k3[2] + k1[2] * k2[2] + k3[2] * k2[2])
            )
        )

        delta3 = (
            c3pp
            * c1p
            * (
                c2m
                * c2pp
                * (-k2[2] ** 2 - k2[2] * k1[2] + k1[2] * k3[2] + k2[2] * k3[2])
                + c2p
                * c2pm
                * (k2[2] ** 2 - k2[2] * k1[2] + k2[2] * k3[2] - k1[2] * k3[2])
            )
        )

        delta2 = 2.0 * c2m * c1p * (k1[2] * k3[2] + k3[2] * k2[2])

        delta2p = 2.0 * c2p * c1p * (k2[2] * k3[2] - k1[2] * k3[2])

        delta1 = 4.0 * k2[2] * k3[2]

        cep3 = delta3 / delta
        ce2 = delta2 / delta
        cep2 = delta2p / delta
        ce1 = delta1 / delta

        ### --- modulus magnetic field in p-polarized mode
        deltam = (
            c3pm
            * c1p
            * (
                c2m * c2pp * (k2gz**2 + k1gz * k3gz + k3gz * k2gz + k1gz * k2gz)
                + c2p * c2pm * (-(k2gz**2) - k1gz * k3gz + k1gz * k2gz + k3gz * k2gz)
            )
        )

        delta3m = (
            c3pp
            * c1p
            * (
                c2m * c2pp * (-(k2gz**2) - k2gz * k1gz + k1gz * k3gz + k2gz * k3gz)
                + c2p * c2pm * (k2gz**2 - k2gz * k1gz + k2gz * k3gz - k1gz * k3gz)
            )
        )

        delta2m = 2.0 * c2m * c1p * (k1gz * k3gz + k3gz * k2gz)

        delta2pm = 2.0 * c2p * c1p * (k2gz * k3gz - k1gz * k3gz)

        delta1m = 4.0 * k2gz * k3gz

        cmagp3 = delta3m / deltam
        cmag2 = delta2m / deltam
        cmagp2 = delta2pm / deltam
        cmag1 = delta1m / deltam

        ### --- Determination of the differents electric and magnetic field
        if z > z_u:
            cphase3 = cmath.exp(1.0j * np.dot(k3, r))
            cphase3p = cmath.exp(1.0j * np.dot(k3p, r))

            if polar == "s":
                Ex = 0.0
                Ey = cphase3 + cep3 * cphase3p
                Ez = 0.0

                Bx = -cphase3 * k3[2] / k0 - (cep3 * cphase3p * k3p[2] / k0)
                By = 0.0
                Bz = cphase3 * k3[0] / k0 + (cep3 * cphase3p * k3p[0] / k0)

            if polar == "p":
                Ex = n3 * (
                    -cphase3 * k3[2] / (eps3 * k0)
                    - cmagp3 * cphase3p * k3p[2] / (eps3 * k0)
                )
                Ey = 0.0
                Ez = n3 * (
                    cphase3 * k3[0] / (eps3 * k0)
                    + cmagp3 * cphase3p * k3p[0] / (eps3 * k0)
                )

                Bx = 0.0
                By = n3 * (-cphase3 - cmagp3 * cphase3p)
                Bz = 0.0

        elif z_d < z < z_u:
            cphase2 = cmath.exp(1.0j * np.dot(k2, r))
            cphase2p = cmath.exp(1.0j * np.dot(k2p, r))

            if polar == "s":
                Ex = 0.0
                Ey = ce2 * cphase2 + cep2 * cphase2p
                Ez = 0.0

                Bx = -ce2 * cphase2 * k2[2] / k0 - cep2 * cphase2p * k2p[2] / k0
                By = 0.0
                Bz = ce2 * cphase2 * k2[0] / k0 + cep2 * cphase2p * k2p[0] / k0

            if polar == "p":
                Ex = n3 * (
                    -cmag2 * cphase2 * k2[2] / (eps2 * k0)
                    - cmagp2 * cphase2p * k2p[2] / (eps2 * k0)
                )
                Ey = 0.0
                Ez = n3 * (
                    cmag2 * cphase2 * k2[0] / (eps2 * k0)
                    + cmagp2 * cphase2p * k2p[0] / (eps2 * k0)
                )

                Bx = 0.0
                By = n3 * (-cmag2 * cphase2 - cmagp2 * cphase2p)
                Bz = 0.0

        else:
            cphase1 = cmath.exp(1.0j * np.dot(k1, r))

            if polar == "s":
                Ex = 0.0
                Ey = ce1 * cphase1
                Ez = 0.0

                Bx = -ce1 * cphase1 * k1[2] / k0
                By = 0.0
                Bz = ce1 * cphase1 * k1[0] / k0

            if polar == "p":
                Ex = n3 * (-cmag1 * cphase1 * k1[2] / (eps1 * k0))
                Ey = 0.0
                # Ez = 1./n3*(cmag1*cphase1*k1[0]/(eps1*k0))  # <-- this was probably wrong
                Ez = n3 * (cmag1 * cphase1 * k1[0] / (eps1 * k0))

                Bx = 0.0
                By = n3 * (-cmag1 * cphase1)
                Bz = 0.0

    else:
        ## -- Different wavevectors
        k1 = np.array(
            [-n1 * k0 * cmath.sin(theta_r), 0.0, n1 * k0 * cmath.cos(theta_r)]
        ).astype(
            np.complex64
        )  ## -- incident wavevector in medium 1 (bottom layer)
        k1p = np.array(
            [-n1 * k0 * cmath.sin(theta_r), 0.0, -n1 * k0 * cmath.cos(theta_r)]
        ).astype(
            np.complex64
        )  ## -- reflected wavevector in medium 1 (bottom layer)
        k2 = np.array(
            [
                -n1 * k0 * cmath.sin(theta_r),
                0.0,
                k0 * cmath.sqrt(eps2 - eps1 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- transmitted wavevector in medium 2 (middle layer)
        k2p = np.array(
            [
                -n1 * k0 * cmath.sin(theta_r),
                0.0,
                -k0 * cmath.sqrt(eps2 - eps1 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- reflected wavevector in medium 2 (middle layer)
        k3 = np.array(
            [
                -n1 * k0 * cmath.sin(theta_r),
                0.0,
                k0 * cmath.sqrt(eps3 - eps1 * cmath.sin(theta_r) ** 2),
            ]
        ).astype(
            np.complex64
        )  ## -- transmitted wavevector in medium 3 (top layer)

        ## -- Phase terms
        c1p = cmath.exp(1.0j * k1[2] * z_d)
        c1m = cmath.exp(-1.0j * k1[2] * z_d)
        c2p = cmath.exp(1.0j * k2[2] * z_d)
        c2m = cmath.exp(-1.0j * k2[2] * z_d)
        c2pp = cmath.exp(1.0j * k2[2] * z_u)
        c2pm = cmath.exp(-1.0j * k2[2] * z_u)
        c3pp = cmath.exp(1.0j * k3[2] * z_u)
        c3pm = cmath.exp(-1.0j * k3[2] * z_u)

        ## -- z - components of the wavevector/eps for magnetic modulus
        k1gz = k1[2] / eps1
        k2gz = k2[2] / eps2
        k3gz = k3[2] / eps3

        ### --- modulus electric field in s-polarized mode
        delta = (
            c3pp
            * c1m
            * (
                c2m
                * c2pp
                * (-k2[2] ** 2 - k3[2] * k1[2] + k2[2] * k3[2] + k2[2] * k1[2])
                + c2p
                * c2pm
                * (k1[2] * k2[2] + k3[2] * k1[2] + k3[2] * k2[2] + k2[2] ** 2)
            )
        )

        delta1 = (
            c3pp
            * c1p
            * (
                c2m
                * c2pp
                * (k2[2] ** 2 - k3[2] * k2[2] + k1[2] * k2[2] - k1[2] * k3[2])
                + c2p
                * c2pm
                * (-k2[2] ** 2 - k3[2] * k2[2] + k1[2] * k3[2] + k1[2] * k2[2])
            )
        )

        delta2 = 2.0 * c3pp * c2pm * (k1[2] * k2[2] + k1[2] * k3[2])

        delta2p = 2.0 * c3pp * c2pp * (k1[2] * k2[2] - k1[2] * k3[2])

        delta3 = 4.0 * k1[2] * k2[2]

        cep1 = delta1 / delta
        ce2 = delta2 / delta
        cep2 = delta2p / delta
        ce3 = delta3 / delta

        ### --- modulus magnetic field in p-polarized mode
        deltam = (
            c3pp
            * c1m
            * (
                c2m * c2pp * (-(k2gz**2) - k3gz * k1gz + k2gz * k3gz + k2gz * k1gz)
                + c2p * c2pm * (k1gz * k2gz + k3gz * k1gz + k3gz * k2gz + k2gz**2)
            )
        )

        delta1m = (
            c3pp
            * c1p
            * (
                c2m * c2pp * (k2gz**2 - k3gz * k2gz + k1gz * k2gz - k1gz * k3gz)
                + c2p * c2pm * (-(k2gz**2) - k3gz * k2gz + k1gz * k3gz + k1gz * k2gz)
            )
        )

        delta2m = 2.0 * c3pp * c2pm * (k1gz * k2gz + k1gz * k3gz)

        delta2pm = 2.0 * c3pp * c2pp * (k1gz * k2gz - k1gz * k3gz)

        delta3m = 4.0 * k1gz * k2gz

        cmagp1 = delta1m / deltam
        cmag2 = delta2m / deltam
        cmagp2 = delta2pm / deltam
        cmag3 = delta3m / deltam

        ### --- Determination of the differents electric and magnetic field
        if z < z_d:
            cphase1 = cmath.exp(1.0j * np.dot(k1, r))
            cphase1p = cmath.exp(1.0j * np.dot(k1p, r))

            if polar == "s":
                Ex = 0.0
                Ey = cphase1 + cep1 * cphase1p
                Ez = 0.0

                Bx = -cphase1 * k1[2] / k0 - (cep1 * cphase1p * k1p[2] / k0)
                By = 0.0
                Bz = cphase1 * k1[0] / k0 + (cep1 * cphase1p * k1p[0] / k0)

            if polar == "p":
                Ex = n1 * (
                    -cphase1 * k1[2] / (eps1 * k0)
                    - cmagp1 * cphase1p * k1p[2] / (eps1 * k0)
                )
                Ey = 0.0
                Ez = n1 * (
                    cphase1 * k1[0] / (eps1 * k0)
                    + cmagp1 * cphase1p * k1p[0] / (eps1 * k0)
                )

                Bx = 0.0
                By = n1 * (-cphase1 - cmagp1 * cphase1p)
                Bz = 0.0

        elif z_d < z < z_u:
            cphase2 = cmath.exp(1.0j * np.dot(k2, r))
            cphase2p = cmath.exp(1.0j * np.dot(k2p, r))

            if polar == "s":
                Ex = 0.0
                Ey = ce2 * cphase2 + cep2 * cphase2p
                Ez = 0.0

                Bx = -ce2 * cphase2 * k2[2] / k0 - cep2 * cphase2p * k2p[2] / k0
                By = 0.0
                Bz = ce2 * cphase2 * k2[0] / k0 + cep2 * cphase2p * k2p[0] / k0

            if polar == "p":
                Ex = n1 * (
                    -cmag2 * cphase2 * k2[2] / (eps2 * k0)
                    - cmagp2 * cphase2p * k2p[2] / (eps2 * k0)
                )
                Ey = 0.0
                Ez = n1 * (
                    cmag2 * cphase2 * k2[0] / (eps2 * k0)
                    + cmagp2 * cphase2p * k2p[0] / (eps2 * k0)
                )

                Bx = 0.0
                By = n1 * (-cmag2 * cphase2 - cmagp2 * cphase2p)
                Bz = 0.0

        else:
            cphase3 = cmath.exp(1.0j * np.dot(k3, r))

            if polar == "s":
                Ex = 0.0
                Ey = ce3 * cphase3
                Ez = 0.0

                Bx = -ce3 * cphase3 * k3[2] / k0
                By = 0.0
                Bz = ce3 * cphase3 * k3[0] / k0

            if polar == "p":
                Ex = n1 * (-cmag3 * cphase3 * k3[2] / (eps3 * k0))
                Ey = 0.0
                # Ez = 1./n1*(cmag3*cphase3*k3[0]/(eps3*k0))  # <-- this was probably wrong
                Ez = n1 * (cmag3 * cphase3 * k3[0] / (eps3 * k0))

                Bx = 0.0
                By = n1 * (-cmag3 * cphase3)
                Bz = 0.0

    return E0 * Ex, E0 * Ey, E0 * Ez, E0 * Bx, E0 * By, E0 * Bz


# =============================================================================
# interface to `_three_layer_pw`
# =============================================================================
def plane_wave(
    pos,
    env_dict,
    wavelength,
    inc_angle=180,
    inc_plane="xz",
    theta=None,
    E_s=0.0,
    E_p=1.0,
    phase_Es=0.0,
    phase=0.0,
    returnField="E",
):
    """generalized incident planewave

    supports oblique angles, arbitrary polarization states, 2 interfaces.
    Default config gives incident plane wave from top (z),
    linear polarized along x, with amplitude 1

    Original code by Ch. Girard, python implementation by C. Majorel

    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]

    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].

    wavelength : float
        Wavelength in nm

    inc_angle : float, default: 180
        incident angle with respect to e_z, in degrees. Default is inc from top.
         - 0deg = along Z (from neg to pos Z)
         - 90deg = along X ['xz'], along Y ['yz'] (from pos to neg)
         - 180deg = along Z  (from pos to neg Z)
         - 270deg = along X ['xz'], along Y ['yz'] (from neg to pos)

    inc_plane : str, default: 'xz'
        plane of incidence, one of ['xz', 'yz']

    theta : float, default: None
        alternative specification for a linear polarization angle.
        If given, this will override `E_s` and `E_p` as well as their respective phases.
        In degrees.
        At normal incidence: 0deg = OX, 90deg = OY.
        If `inc_plane`=='xz': 0deg --> p-polarization; 90deg --> s-polarization
        (inverse for `inc_plane`=='yz')


    E_s, E_p : float, default: 0.0, 1.0
        Apmplitudes of s-polarized and p-polarized plane wave components.
        At 0 / 180 degrees incident angle (normal incindence), 'p' is
        polarized along x, 's' along y. Then, at 90deg 'p' is along z.

    phase_Es : float, default: 0.0
        additional phase for E_s component (in rad).
        Can be used to generate elliptic polarization.
        For instance, left circular polarization (LCP) can be obtained with:
        E_s=np.sqrt(0.5), E_p=np.sqrt(0.5), phase_Es=-np.pi/2.
        RCP: phase_Es=+np.pi/2.

    phase : float, default: 0.0
        additional absolute phase for entire plane wave (in rad).

    returnField : str, default: 'E'
        if 'E': returns electric field; if 'B' or 'H': magnetic field

    Returns
    -------
      E0 (B0):       Complex E-(B-)Field at each dipole position as
                    list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]
    """
    if (
        "eps1" not in env_dict.keys()
        or "eps2" not in env_dict.keys()
        or "eps3" not in env_dict.keys()
        or "spacing" not in env_dict.keys()
    ):
        raise ValueError("`env_dict` must contain ['eps1', 'eps2', 'eps3', 'spacing']")

    cn1 = env_dict["eps1"] ** 0.5
    cn2 = env_dict["eps2"] ** 0.5
    cn3 = env_dict["eps3"] ** 0.5
    spacing = np.float32(env_dict["spacing"].real)
    z_d = 0  # position of lower interface

    ## -- convert angles 90 and 270 close to horizontal angles to avoid divergence
    if inc_angle in [-90, 90, -270, 270] and (cn1 != cn2 or cn2 != cn3):
        warnings.warn(
            "Using interface with horizontal angle of incidence!"
            + "Please make sure if horizontal incidence makes sense in presence of an interface."
        )
    if inc_angle in [-90, 90]:
        inc_angle += 0.05
    if inc_angle in [-270, 270]:
        inc_angle -= 0.05

    Ex = np.asfortranarray(np.zeros(len(pos)), dtype=DTYPE_C)
    Ey = np.asfortranarray(np.zeros(len(pos)), dtype=DTYPE_C)
    Ez = np.asfortranarray(np.zeros(len(pos)), dtype=DTYPE_C)

    if theta is not None:
        if inc_angle not in [0, 180]:
            warnings.warn(
                "non-normal incident angle, to avoid ambiguities, "
                + "the polarization of a plane wave should not be "
                + "defined via the `theta` keyword."
            )
        E_p = 1.0 * np.cos(theta * np.pi / 180.0)
        E_s = 1.0 * np.sin(theta * np.pi / 180.0)

    ## normalize incident amplitude by refractive index
    # if abs(inc_angle)<90 and np.abs(cn1)!=1: # inc from zone n1
    #     E_s = E_s / cn1**0.5
    #     E_p = E_p / cn1**0.5
    # elif abs(inc_angle)>=90 and np.abs(cn3)!=1: # inc from zone n3
    #     E_s = E_s / cn3**0.5
    #     E_p = E_p / cn3**0.5

    for i, R in enumerate(pos):
        if inc_plane.lower() in ["yz", "zy"]:
            y, x, z = R
        else:
            x, y, z = R

        ex_s, ey_s, ez_s, bx_s, by_s, bz_s = _three_layer_pw(
            wavelength, inc_angle, "s", z_d, spacing, cn1, cn2, cn3, x, y, z, E0=E_s
        )

        ex_p, ey_p, ez_p, bx_p, by_p, bz_p = _three_layer_pw(
            wavelength, inc_angle, "p", z_d, spacing, cn1, cn2, cn3, x, y, z, E0=E_p
        )

        ## additional phases:
        ex_p = ex_p * np.exp(1j * phase)
        ey_p = ey_p * np.exp(1j * phase)
        ez_p = ez_p * np.exp(1j * phase)
        bx_p = bx_p * np.exp(1j * phase)
        by_p = by_p * np.exp(1j * phase)
        bz_p = bz_p * np.exp(1j * phase)
        ex_s = ex_s * np.exp(1j * (phase_Es + phase))
        ey_s = ey_s * np.exp(1j * (phase_Es + phase))
        ez_s = ez_s * np.exp(1j * (phase_Es + phase))
        bx_s = bx_s * np.exp(1j * (phase_Es + phase))
        by_s = by_s * np.exp(1j * (phase_Es + phase))
        bz_s = bz_s * np.exp(1j * (phase_Es + phase))

        ## optional scattering plane modification
        if inc_plane.lower() in ["yz", "zy"]:
            ex_s, ey_s, ez_s = -ey_s, -ex_s, -ez_s
            bx_s, by_s, bz_s = by_s, bx_s, bz_s

            ex_p, ey_p, ez_p = -ey_p, -ex_p, -ez_p
            bx_p, by_p, bz_p = by_p, bx_p, bz_p

        if returnField.lower() == "e":
            Ex[i], Ey[i], Ez[i] = ex_s + ex_p, ey_s + ey_p, ez_s + ez_p
        else:
            Ex[i], Ey[i], Ez[i] = bx_s + bx_p, by_s + by_p, bz_s + bz_p

    Evec = np.transpose([Ex, Ey, Ez])
    return Evec


def gaussian(
    pos,
    env_dict,
    wavelength,
    theta=None,
    polarization_state=None,
    xSpot=0.0,
    ySpot=0.0,
    zSpot=0.0,
    NA=-1.0,
    spotsize=-1.0,
    kSign=-1.0,
    paraxial=False,
    phase=0.0,
    E0=complex(1, 0),
    returnField="E",
):
    """Simple, normal incident (along Z) focused Gaussian Beam Field

    For more advanced options, use fields from `focused_beams` module.

    obligatory "einKwargs" are one of 'theta' or 'polarization_state' and
    one of 'NA' or 'spotsize'

    polarization is defined by one of the two kwargs:
     - theta: linear polarization angle in degrees, theta=0 --> along X.
              Amplitude = 1 for both, E and B
     - polarization_state. tuple (E0x, E0y, Dphi, phi): manually
              define x and y amplitudes, phase difference and
              absolute phase of plane wave.


    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]

    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].

    wavelength : float
        Wavelength in nm

    theta : float, default: None
        either 'theta' or 'polarization_state' must be given.
        linear polarization angle in degrees, 0deg = 0X direction

    polarization_state : 4-tuple of float, default: None
        either 'theta' or 'polarization_state' must be given.
        polarization state with field amplitudes and phases, tuple of 4 float:
        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference
        between X and Y components (in rad), absolute phase of plane wave (in rad).
        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
        Note that this means the handedness depends on the propagation direction (*kSign*)!
            Dphi :
                - positive: left hand rotating polarization
                - negative: right hand rotating polarization
                - example: left circular pol. with (kSign=-1, E0x=1, E0y=1, Dphi=np.pi/2., phi=0)

      xSpot,ySpot,zSpot : float, default: 0,0,0
          x/y/z coordinates of focal point

      NA : float
          Numerical aperture to calculate beamwaist

      spotsize : float (optional)
          Gaussian beamwaist (overrides "NA")

      kSign : float, default: -1
          Direction of Beam. -1: top to Bottom, 1 Bottom to top

      paraxial : bool, default: False
          Use paraxial Gaussian beam: No longitudinal fields.
          If "False", longitudinal components are obtained using Maxwell
          equation div(E)=0 as condition

      phase : float, default: 0
          additional phase of the beam, in degrees

      E0 : complex or function, default: complex(1,0)
          Either complex value or function of r=(x, y) (normalized to units of waist!).
          In case of a function, it needs to return the  complex amplitude at
          the given position relative to beam axis (pos. in units of waist).

      returnField : str, default: 'E'
          if 'E': returns electric field; if 'B' or 'H': magnetic field

    Returns
    -------
      E0:       Complex E-Field at each dipole position


    Notes
    -----
     - paraxial correction :
         see: Novotny & Hecht. "Principles of nano-optics". Cambridge University Press (2006)


    """
    if (theta is None and polarization_state is None) or (
        theta is not None and polarization_state is not None
    ):
        raise ValueError(
            "exactly one argument of 'theta' and 'polarization_state' must be given."
        )
    if theta is not None:
        polarization_state = (
            1.0 * np.cos(theta * np.pi / 180.0),
            1.0 * np.sin(theta * np.pi / 180.0),
            0,
            0,
        )

    xm, ym, zm = np.transpose(pos)

    if "eps_env" in env_dict.keys():
        cn1 = cn2 = env_dict["eps_env"] ** 0.5
    else:
        cn1 = env_dict["eps1"] ** 0.5
        cn2 = env_dict["eps2"] ** 0.5
        cn3 = env_dict["eps3"] ** 0.5
        # spacing = env_dict['spacing']**0.5
        if cn1 != cn2 or cn2 != cn3:
            warnings.warn(
                "`gaussian` only supports vacuum environment so far. "
                + "A simulation with interface might not yield correct results."
            )

    Ex = np.asfortranarray(np.zeros(len(xm)), dtype=DTYPE_C)
    Ey = np.asfortranarray(np.zeros(len(xm)), dtype=DTYPE_C)
    Ez = np.asfortranarray(np.zeros(len(xm)), dtype=DTYPE_C)

    ## beamwaist
    if spotsize == NA == -1:
        raise ValueError("Focused Beam Error! Either spotsize or NA must be given.")
    elif spotsize == -1:
        w0 = 2 * wavelength / (NA * np.pi)
    else:
        w0 = spotsize

    ## waist, curvature and gouy-phase
    def w(z, zR, w0):
        return w0 * np.sqrt(1 + (z / zR) ** 2)

    def R(z, zR):
        return z * (1 + (zR / z) ** 2)

    def gouy(z, zR):
        return np.arctan2(z, zR)

    ## constant parameters
    k = kSign * cn2 * (2 * np.pi / wavelength)  # incidence from positive Z
    zR = np.pi * w0**2 / wavelength

    r2 = (xm - xSpot) ** 2 + (ym - ySpot) ** 2
    z = zm - zSpot

    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]

    ##  --------- Electric field ---------
    r = np.transpose([xm - xSpot, ym - ySpot])
    waist = w(z, zR, w0)
    if isinstance(E0, complex):
        _E0 = E0
    elif isinstance(E0, types.FunctionType):
        _E0 = E0(r / waist[:, None])
    else:
        raise Exception(
            "Wrong type for complex amplitude `E0`. Must be `complex` or `function`."
        )
    E = (
        _E0
        * (
            (w0 / waist)
            * np.exp(-r2 / waist**2)
            * np.exp(1j * (k * z + k * r2 / (2 * R(z, zR)) - gouy(z, zR)))
        )
        * np.exp(1j * phase * np.pi / 180.0)
    )

    abs_phase = np.exp(1j * Aphi)  # add an absolute phase
    if returnField.lower() == "e":
        Ex = E * E0x * abs_phase
        Ey = E * E0y * np.exp(1j * Dphi) * abs_phase
    else:
        ##  --------- Magnetic field ---------
        Ex = E * E0y * np.exp(1j * Dphi) * abs_phase
        Ey = -1 * E * E0x * abs_phase

    ## obtained longitudinal component using condition div(E)==0
    if paraxial:
        Ez = np.zeros(len(E))  # <-- paraxial gaussian beam: No longitudinal E-component
    else:
        Ez = (-1j * 2 / (k * w(z, zR, w0) ** 2)) * (
            (xm - xSpot) * Ex + (ym - ySpot) * Ey
        )

    Evec = np.transpose([Ex, Ey, Ez]).astype(DTYPE_C)
    return Evec


def dipole_electric(
    pos,
    env_dict,
    wavelength,
    x0,
    y0,
    z0,
    mx,
    my,
    mz,
    returnField="E",
    R_farfield_approx=-1,
):
    """field emitted by an electric dipole at (x0,y0,z0) with complex amplitude (mx,my,mz)

    mandatory kwargs along with `wavelength` are: `x0`, `y0`, `z0`, `mx`, `my`, `mz`

    To take into account a dielectric interface, `dipole_electric` uses a
    mirror-charge approximation in the (quasistatic) near-field and an
    asymptotic approximation for the far-field. Can handle only a single interface
    (hence cases with n1 != n2 = n3).


    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]

    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain either "eps_env" or ["eps1", "eps2"].

    wavelength : float
        Wavelength in nm

    x0,y0,z0 : float
        x/y/z coordinates of electric dipole position

    mx,my,mz : float
        x/y/z amplitude of elec. dipole vector

    returnField : str, default: 'E'
        if 'E': returns electric field; if 'B' or 'H': magnetic field

    R_farfield_approx : float, default: -1
        optional emitter-observer distance (in nm) starting from which an asymptotic
        farfield approximation will be used (to be used with caution!).
        `-1`: Do not use far-field approximation.

    Returns
    -------
    E0/H0:   Complex field at each position ( (Ex,Ey,Ez)-tuples )

    Notes
    -----

    for free-space propagators, see e.g.
    G. S. Agarwal, *Phys. Rev. A*, 11(230), (1975), Eqs. (4.5)/(4.6)

    """
    from pyGDM2.propagators import Gtot_EE_123 as GEE
    from pyGDM2.propagators import Gs_EE_asymptotic as GEE_ff
    from pyGDM2.propagators import G_HE_123 as GHE

    if "eps_env" in env_dict.keys():
        eps1 = eps2 = eps3 = env_dict["eps_env"]
        spacing = 5000
    else:
        eps1 = env_dict["eps1"]
        eps2 = env_dict["eps2"]
        eps3 = env_dict["eps3"]
        spacing = np.float32(env_dict["spacing"].real)
        if eps2 != eps3:
            warnings.warn(
                "dipole_electric only supports a single interface "
                + "(between `n1`/`n2`). "
                + "The simulation might not be a good approximation."
            )

    R1 = np.array([x0, y0, z0])  # emitter location
    p = np.array([mx, my, mz])  # emitter dipole moment

    Ex = np.zeros(len(pos), dtype=DTYPE_C)
    Ey = np.zeros(len(pos), dtype=DTYPE_C)
    Ez = np.zeros(len(pos), dtype=DTYPE_C)

    ## calc propagator
    for i, R2 in enumerate(pos):
        if returnField.lower() == "e":
            ## --- emitted electric field
            if np.linalg.norm(R2 - R1) <= R_farfield_approx or R_farfield_approx == -1:
                ## mirror-charge NF approximation
                xx, yy, zz, xy, xz, yx, yz, zx, zy = GEE(
                    R1, R2, wavelength, eps1, eps2, eps3, spacing
                )
            else:
                ## asymptotic farfield approximation:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = GEE_ff(
                    R1, R2, wavelength, eps1, eps2, eps3, spacing
                )

        else:
            ## --- emitted magnetic field
            xx, yy, zz, xy, xz, yx, yz, zx, zy = GHE(
                R1, R2, wavelength, eps1, eps2, eps3, spacing
            )

        ## propagate the dipole
        G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]])
        E = np.matmul(G, p)

        Ex[i] = E[0]
        Ey[i] = E[1]
        Ez[i] = E[2]

    return np.transpose([Ex, Ey, Ez])


def dipole_magnetic(
    pos,
    env_dict,
    wavelength,
    x0,
    y0,
    z0,
    mx,
    my,
    mz,
    returnField="E",
    R_farfield_approx=-1,
):
    """field emitted by a magnetic dipole at (x0,y0,z0) with complex amplitude (mx,my,mz)

    mandatory kwargs along with `wavelength` are: `x0`, `y0`, `z0`, `mx`, `my`, `mz`

    To take into account a dielectric interface, `dipole_magnetic` uses a
    mirror-charge approximation in the (quasistatic) near-field and an
    asymptotic approximation for the far-field. Can handle only a single interface
    (hence cases with n1 != n2 = n3).


    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]

    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain either "eps_env" or ["eps1", "eps2"].

    wavelength : float
        Wavelength in nm

    x0,y0,z0 : float
        x/y/z coordinates of electric dipole position

    mx,my,mz : float
        x/y/z amplitude of elec. dipole vector

    returnField : str, default: 'E'
        if 'E': returns electric field; if 'B' or 'H': magnetic field

    R_farfield_approx : float, default: -1
        optional emitter-observer distance (in nm) starting from which an asymptotic
        farfield approximation will be used (to be used with caution!).
        `-1`: Do not use far-field approximation.

    Returns
    -------
    E0/H0:   Complex field at each position ( (Ex,Ey,Ez)-tuples )

    Notes
    -----

    for free-space propagators, see e.g.
    G. S. Agarwal, *Phys. Rev. A*, 11(230), (1975), Eqs. (4.5)/(4.6)

    """
    from pyGDM2.propagators import Gtot_EE_123 as GEE
    from pyGDM2.propagators import Gs_EE_asymptotic as GEE_ff
    from pyGDM2.propagators import G_HE_123 as GHE

    if "eps_env" in env_dict.keys():
        eps1 = eps2 = eps3 = env_dict["eps_env"]
        spacing = 5000
    else:
        eps1 = env_dict["eps1"]
        eps2 = env_dict["eps2"]
        eps3 = env_dict["eps3"]
        spacing = np.float32(env_dict["spacing"].real)
        if eps2 != eps3:
            warnings.warn(
                "dipole_electric only supports a single interface "
                + "(between `n1`/`n2`). "
                + "The simulation might not be a good approximation."
            )

    R1 = np.array([x0, y0, z0])  # emitter location
    p = np.array([mx, my, mz])  # emitter dipole moment

    Ex = np.zeros(len(pos), dtype=DTYPE_C)
    Ey = np.zeros(len(pos), dtype=DTYPE_C)
    Ez = np.zeros(len(pos), dtype=DTYPE_C)

    ## calc propagator
    for i, R2 in enumerate(pos):
        if returnField.lower() == "e":
            ## --- emitted electric field
            ## GEH(R1, R2) = GHE(R2, R1)
            xx, yy, zz, xy, xz, yx, yz, zx, zy = GHE(
                R2, R1, wavelength, eps1, eps2, eps3, spacing
            )

        else:
            ## --- emitted magnetic field
            ## GEE(R1, R2) = GHH(R1, R2)
            if np.linalg.norm(R2 - R1) <= R_farfield_approx or R_farfield_approx == -1:
                ## mirror-charge NF approximation
                xx, yy, zz, xy, xz, yx, yz, zx, zy = GEE(
                    R1, R2, wavelength, eps1, eps2, eps3, spacing
                )
            else:
                ## asymptotic farfield approximation:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = GEE_ff(
                    R1, R2, wavelength, eps1, eps2, eps3, spacing
                )

        ## propagate the dipole
        G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]])
        E = np.matmul(G, p)

        Ex[i] = E[0]
        Ey[i] = E[1]
        Ez[i] = E[2]

    return np.transpose([Ex, Ey, Ez])


if __name__ == "__main__":
    pass



# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
pyGDM core routines

"""
from __future__ import print_function
from __future__ import absolute_import

import numpy as np

import gc
import warnings
import copy
import time

import numba

from pyGDM2.propagators import propagators



#==============================================================================
# GLOBAL PARAMETERS
#==============================================================================




#==============================================================================
# EXCEPTIONS
#==============================================================================





# =============================================================================
# Deprecation Test
# =============================================================================
def _check_struct_type(struct):
    return struct
    # ## deprecation test
    # from pyGDM2 import structures
    # if type(struct) == structures.struct:
    #     warnings.warn("deprecated `structs` class is used. Automatically " +
    #                   "overloading with the new `pyGDM2.structures.struct_py` class.")
        
    #     n1, n2, n3 = struct.n1_material, struct.n2_material, struct.n3_material
    #     spacing = struct.spacing
    #     norm = struct.normalization
        
    #     new_struct = structures.struct_py(struct.step, struct.geometry, struct.material,
    #                                       n1=n1, n2=n2, n3=n3, 
    #                                       normalization=norm, spacing=spacing,
    #                                       auto_shift_structure=False, 
    #                                       check_geometry_consistency=False)
    #     return new_struct
    # else:
    #     return struct
    
    
def _check_efield_type(efield):
    pass
        

#==============================================================================
# Simulation container class
#==============================================================================
class simulation(object):
    """
    Bases: object
    
    Main GDM simulation container class - pure-python API
    
    Defines a linear GDM simulation. Contains information on: 
        - *struct* : :class:`.structures.struct`:
            - the geometry of the nano structure 
            - its dielectric function
            
        - *efield* : :class:`.fields.efield`
            - the incident field and the wavelenghts 
            - possibly further incident field configuration 
            
        - *dyads* : :class:`.propagators.DyadsBaseClass`
            - class derived from :class:`.propagators.DyadsBaseClass`
            - contains set of Green's tensors, self-terms and related functions
            - contains environment definition and polarizability calculation
    
    Parameters
    ----------
    
    struct : :class:`.structures.struct`
        structure object
        
    efield : :class:`.fields.efield`
        fundamental field
        
    dyads : :class:`.propagators.DyadsBaseClass`
        set of Green's tensors, polarizabilities, environment info
        
    dtype : (optional) `str`, default: 'f'
        precision of simulation
    
    """
    __name__ = "pygdm2 simulation"
    
    
    def __init__(self, struct, efield, dyads=None, dtype='f', verbose=True):
        """Initialization"""
        
        ## --- struct includes: geometry, material(s), environment
        struct = _check_struct_type(struct)  # compatibility: check if new `struct_py`-API is used
        self.struct = struct
        
        ## --- efield includes: field_generator function, wavelengths, optional kwargs
        _check_efield_type(efield)  # compatibility: check if new `fields`-API is used
        self.efield = efield
        
        ## --- dyads includes: polarizabilities, self-terms, Green dyads, coupling-matrix setup
        if dyads is None:
            warnings.warn('No `dyads` instance specified. Falling back to `propagators.DyadsQuasistatic123`. ' +
                          'This assumes environment configuration through "struct". ' +
                          'This behavior might change in a future release.')
            dyads = propagators.DyadsQuasistatic123()
        self.dyads = dyads
        
        ## Ensure environment definition for 1-2-3 dyads. 
        ## This will be removed in a future release
        try:
            ## if not in 'dyads', set n1, n2, n3 and spacing from 'struct'
            self.dyads._legacyStructCompatibility(self.struct)
        except AttributeError:
            pass
        
        ## --- precision
        if dtype in ['f', 'F']:
            self.dtypef = np.float32
            self.dtypec = np.complex64
        elif dtype in ['d', 'D']:
            self.dtypef = np.float64
            self.dtypec = np.complex128
        self.efield.setDtype(self.dtypef, self.dtypec)
        self.struct.setDtype(self.dtypef, self.dtypec)
        
        ## --- initialize unevaluated data to `None` (postprocessing data)
        self.E = None   # placeholder: scattered E-fields inside structure
        self.H = None   # placeholder: scattered H-fields inside structure
        self.S_P = None # legcay-compatibility: placeholder for decay-rate tensors
    
        ## check consistency of structure / efield with Green's tensor framework
        self.dyads.exceptionHandling(struct=self.struct, efield=self.efield)
        
    
    def scatter(self, **kwargs):
        """wrapper to :func:`.scatter`"""
        return scatter(self, **kwargs)
    
    
    def copy(self):
        import copy
        return copy.deepcopy(self)
    
    
    def __repr__(self):
        from pyGDM2 import tools
        return tools.print_sim_info(self, prnt=False)
    
    def __add__(self, other):
        if type(other) == simulation:
            ## add another simulation: Try combining both
            from pyGDM2 import tools
            return tools.combine_simulations([self, other])
        else:
            raise ValueError('Unknown addition.')
    
    def __eq__(self, other):
        ## identical geometries, identical field_kwargs keys, wavelengths, 
        ##     material, n1
        if (len(self.struct.geometry) == len(other.struct.geometry) and
               len(self.efield.wavelengths) == len(other.efield.wavelengths) and
               len(self.efield.kwargs.keys()) == len(other.efield.kwargs.keys())):
            if (np.sum(np.abs(self.struct.geometry - other.struct.geometry))==0 and 
                    np.sum(np.abs(self.efield.wavelengths - other.efield.wavelengths))==0 and 
                    self.efield.kwargs.keys() == other.efield.kwargs.keys() and 
                    self.struct.material[0].__name__ == other.struct.material[0].__name__ and 
                    self.dyads.n1_material.__name__ == other.dyads.n1_material.__name__ and 
                    self.dyads.n2_material.__name__ == other.dyads.n2_material.__name__ and 
                    self.dyads.n3_material.__name__ == other.dyads.n3_material.__name__ and 
                    self.dyads.spacing == other.dyads.spacing and 
                    self.struct.step == other.struct.step):
                return True
            else:
                return False
        else:
            return False




#==============================================================================
# scattered fields
#==============================================================================
def scatter(sim, method='lu', calc_H=False, verbose=True, callback=None, **kwargs):
    """Perform a linear scattering GDM simulation
    
    Calculate the electric field distribution inside a nano-structure.
    Optionally calculate also the internal magnetic field.
    
    Parameters
    ----------
    sim : :class:`.simulation`
        simulation description
    
    method : string, default: "lu"
        inversion method. One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
         - "lu" LU-decomposition (`scipy.linalg.lu_factor`)
         - "scipyinv" scipy default inversion (`scipy.linalg.inv`)
         - "numpyinv" numpy inversion (`np.linalg.inv`, if numpy compiled accordingly: LAPACK's `dgesv`)
         - "cupy" uses CUDA GPU via `cupy`
         - "cuda" (equivalent to "cupy")
    
    calc_H : bool, default: False
        if True, calculate also the internal magnetic field H. This has an 
        impact on performance and memory requirement, so it is deactivated by 
        default.
    
    verbose : bool, default False
        Print timing info to stdout
        
    callback : func, default: None
        optional callback function, which is called after each wavelength.
        Passes a dict to `callback` containing current wavelength, 
        simulation and timing info. 
        Available dict_keys: ['i_wl', 'wavelength', 'sim', 't_inverse', 't_repropa'].
        Callback needs to return `True` for simulation to continue. 
        If callback returns `False` the simulation will be canceled, and `scatter`
        returns the simulation results until the moment when it was interrupted.
    
      
    Returns
    -------
        1, if finished succesfully
        
        -1, if canceled
    
    
    Notes
    -----
        The complex electric fields inside the structure are also copied into the 
        :class:`.simulation` instance as attribute `simulation.E`
    
        For details on the concept of the generalized propagator, see e.g.:
        Martin, O. J. F. & Girard, C. & Dereux, A. **Generalized Field Propagator 
        for Electromagnetic Scattering and Light Confinement.**
        Phys. Rev. Lett. 74, 526–529 (1995).
    
        The `scipy` solvers (like 'lu', 'ilu' and 'cg') run parallelized if BLAS
        is compiled with multithreading support. See:
        http://scipy-cookbook.readthedocs.io/items/ParallelProgramming.html#use-parallel-primitives
        (last called 05/2020)
      
        To limit the number of threads for the multi-threaded parallelized parts, 
        you might do something like explained in: 
        https://stackoverflow.com/questions/29559338/set-max-number-of-threads-at-runtime-on-numpy-openblas
        (website called 05/2020).
        Or use the `threadpoolctl` package (example for 4 threads:):
          >>> from threadpoolctl import threadpool_limits
          >>> threadpool_limits(limits=4, user_api='blas')
        
    """
# =============================================================================
#     Exception handling
# =============================================================================
    if method.lower() == "cuda":
        method = "cupy"
        
    if method.lower() not in ["numpyinv", "scipyinv", "lu", "cupy"]:
        raise ValueError('Error: Unknown solving method. Must be one of' +
                         ' ["lu", "numpyinv", "scipyinv", "cuda", "cupy"].')
    
    if method.lower() == 'lu' and calc_H:
        warnings.warn("calculating 'H' is not efficient with LU decomposition. " +
                      "Consider using `scipyinv` instead.")
    
    if len(kwargs)!=0:
        warnings.warn("Following function-less kwargs were ignored: {}".format([a for a in kwargs]))
        
    
# =============================================================================
#     iterate wavelengths
# =============================================================================
    field_generator = sim.efield.field_generator
    wavelengths = sim.efield.wavelengths
    
    scattered_fields_E = []
    scattered_fields_H = []
    for i_wl, wavelength in enumerate(wavelengths):
        
# =============================================================================
#     get generalized propagator
# =============================================================================
        t0 = time.time()
        K = get_general_propagator(sim=sim, method=method, 
                                      wavelength=wavelength, calc_H=calc_H,
                                      verbose=verbose)
        t1 = time.time()
        
#==============================================================================
#    each wl: Incident field evaluation
#==============================================================================
        ## --- At fixed wavelength: Use generalized propagator on all incident field parameters
        def generalized_propagator_operation(field_kwargs, K, sim, wavelength):
            env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
            ## --- optional: E0
            E0 = field_generator(sim.struct.geometry, env_dict, wavelength, **field_kwargs)
            E0_supervec = _fieldListToSuperVector(E0)
            
            ## --- optional: H0
            if calc_H:
                field_kwargs_H = copy.deepcopy(field_kwargs)
                field_kwargs_H["returnField"] = "H"
                H0 = field_generator(sim.struct.geometry, env_dict, wavelength, **field_kwargs_H)
                H0_supervec = _fieldListToSuperVector(H0)
                E0_supervec = np.concatenate([E0_supervec, H0_supervec])
                E = np.dot(K, E0_supervec)
                H = _superVectorToFieldList(E[3*len(sim.struct.geometry):])
            
            ## --- generalized propagator times incident field:
            elif method.lower() == 'lu' and not calc_H:
                import scipy.linalg as la
                E = la.lu_solve(K, E0_supervec)
            elif method.lower() == 'cupy' and type(K) != np.ndarray:
                ## --- cupy GPU matrix vector multiplication
                import cupy as cp
                E0_supervec_gpu = cp.array(np.ascontiguousarray(E0_supervec))  # move vec to device memory
                E_gpu = cp.dot(K, E0_supervec_gpu)
                E = E_gpu.get()
            else:
                E = np.dot(K, E0_supervec)
            
            E = _superVectorToFieldList(E[:3*len(sim.struct.geometry)])
            kwargs_final = copy.deepcopy(field_kwargs)
            kwargs_final["wavelength"] = wavelength
            
            ## --- optional: H
            if calc_H:
                return dict(kw=kwargs_final, E=E, H=H)
            else:
                return dict(kw=kwargs_final, E=E, H=None)
        
        ## --- loop over incident field configurations
        for field_kwargs in sim.efield.kwargs_permutations:
            scat_results = generalized_propagator_operation(field_kwargs, K, sim, wavelength)
            scattered_fields_E.append([scat_results['kw'], scat_results['E']]) 
            scattered_fields_H.append([scat_results['kw'], scat_results['H']])
            sim.E = scattered_fields_E
            if calc_H:
                sim.H = scattered_fields_H
                
        if verbose: print("repropa.: {:.1f}ms ({} field configs), tot: {:.1f}ms".format(
                                        1000.*(time.time()-t1), 
                                        len(sim.efield.kwargs_permutations), 
                                        1000.*(time.time()-t0)))
# =============================================================================
#       if applicable: call callback
# =============================================================================
        if callback is not None:
            cb_continue = callback(dict(
                i_wl=i_wl, wavelength=wavelength, sim=sim,
                t_inverse=1000.*(t1-t0), t_repropa=1000.*(time.time()-t1))
                                  )
            if not cb_continue:
                return -1
        
        ## free RAM for next iteration
        del K
        # gc.collect()
    
    return 1



## !!! Not working due to numba conflicting with multiprocessing
def _scatter_multiprocessing(sim, calc_H=False, verbose=False, scatter_verbose=False, 
                            N_process=-1, N_thread=1, **kwargs):
    """multiprocessing parallel execution of :func:`.scatter` for parallel calculation of spectra
    
    *requires:* **threadpoolctl** (for setting number of numpy parallel threads)
    
    Parameters
    ----------
    sim : :class:`.simulation` 
        simulation description
    
    calc_H : bool, default: False
        if True, calculate also the internal magnetic field H.
    
    verbose, scatter_verbose: bool, default: False
        turns on some mpi-routine info printing, respectively controls verbose setting for :func:`.scatter`
    
    N_process : int, default: -1
        optional number of parallel processes to use for scattering evaluation. 
        '-1': use number of available CPUs
    
    N_thread : int, default: 1
        optional number of CPUS to allocate for each thread (each wavelength). 
        '-1': use number of available CPUs
    
    **kwargs : 
        all kwargs are passed to :func:`.scatter`


    Notes 
    -----
    - The product of `N_process` x `N_thread` shoould not exceed the number of CPUs.
      *Overloading the CPUs will usually result in **decreased** calculation speed!*
    
    - The memory requirements are `N_process` times higher as for single execution!
    
    - see :func:`.scatter` for main documentation
    
    """
    from pyGDM2 import fields
    from pyGDM2.tools import _get_nr_processes
    import multiprocessing
    
    if N_process == -1:
        N_process = _get_nr_processes()
    if N_thread == -1:
        N_thread = _get_nr_processes()
    if verbose: 
        print("")
        print("number of processes:             ", N_process)
        print("number of threads / wavelength:  ", N_thread)
        print("number of wavelengths / process: ", int(np.ceil(len(sim.efield.wavelengths) / float(N_process))))
        print("")
    
    from threadpoolctl import threadpool_limits
    import numba
    threadpool_limits(limits=N_thread, user_api='blas')
    numba.set_num_threads(N_thread)
    
    
    
    ## --- create list of jobs and split in equal parts depending on `nprocs`
    def split(jobs, nprocs):
        return [jobs[i::nprocs] for i in range(nprocs)]
    
    if verbose: print("Generating jobs... ", end='')
    jobs = []
    for i, wl in enumerate(sim.efield.wavelengths):
        ## --- generate simulation objects for each individual wavelength
        _sim = copy.deepcopy(sim)
        _efield = fields.efield(_sim.efield.field_generator, [wl], _sim.efield.kwargs)
        _efield.setDtype(_sim.dtypef, _sim.dtypec)
        _sim.efield = copy.deepcopy(_efield)
        jobs.append(_sim)
    # jobs = split(jobs_all, N_process)
    if len(jobs)%N_process != 0:
        warnings.warn("Efficiency warning: Number of wavelengths ({}) ".format(
                                            len(sim.efield.wavelengths)) + 
                      "not divisable by Nr of processes ({})!".format(N_process))
    if verbose: print("Done.")
    
    
    ## --- evaluation of `scatter` in parallel for each wavelength
    if verbose: print("Running on {} processes... ".format(N_process), end='')
    t0 = time.time()
    global worker_wrapper
    def worker_wrapper(job):
        return scatter(job, calc_H=calc_H, verbose=scatter_verbose, **kwargs)
    
    ## --- multiprocessing on N_cpu tasks
    with multiprocessing.Pool(N_process) as p:
        _ = p.map(worker_wrapper, jobs)
    
    ## --- gathering results
    if verbose: print("All simulations done. Recombining... ", end='')
    resultsE = []
    resultsH = []
    for job in jobs:
        scattered_Efield = job.E
        for _scatf in scattered_Efield:
            resultsE.append(_scatf)
        if calc_H:
            scattered_Hfield = job.H
            for _scatf in scattered_Hfield:
                resultsH.append(_scatf)
    
    ## --- recombine data and sort by wavelength
    resultsE = [i for temp in resultsE for i in temp]
    resultsE = sorted(resultsE, key=lambda k: k[0]['wavelength'])
    sim.E = resultsE
    if calc_H:
        resultsH = [i for temp in resultsH for i in temp]
        resultsH = sorted(resultsH, key=lambda k: k[0]['wavelength'])
        sim.H = resultsH
    if verbose: print("Done.")
    
    
    if verbose:
        print('calculated {} wavelengths in {:.2f}s (working on {} processes)'.format(
                        len(sim.efield.wavelengths), time.time()-t0, N_process))
    
    return 1
    
        
def scatter_mpi(sim, calc_H=False, verbose=False, scatter_verbose=False, **kwargs):
    """MPI wrapper to :func:`.scatter` for embarrassingly parallel calculation of spectra
    
    *requires:* **mpi4py**
    
    run with "mpirun -n X python scriptname.py", where `X` is the number of 
    parallel processes (ideally an integer divisor of the number of wavelengths)
    
    
    Parameters
    ----------
    sim : :class:`.simulation` 
        simulation description
    
    calc_H : bool, default: False
        if True, calculate also the internal magnetic field H.
    
    verbose, scatter_verbose: bool, default: False
        turns on some mpi-routine info printing, respectively controls verbose setting for :func:`.scatter`
    
    **kwargs : 
        all kwargs are passed to :func:`.scatter`


    Notes 
    -----
    - On single machines it is usually easier to install `scipy` 
      compiled with parallel BLAS (parallel LU / CG routines). Usually the
      parallel BLAS is installed by default. Using both
      parallelisation techniques simultaneously requires proper
      configuration of threads / process numbers. 
      *Overloading the CPUs will usually result in **decreased** calculation speed!*
    
    - see :func:`.scatter` for main documentation
    
    """
    from pyGDM2 import fields
    from mpi4py import MPI
    
    comm = MPI.COMM_WORLD
    nprocs, rank = comm.Get_size(), comm.Get_rank()
    
    
    ## --- create list of jobs and split in equal parts depending on `nprocs`
    def split(jobs, nprocs):
        return [jobs[i::nprocs] for i in range(nprocs)]
    
    if comm.rank == 0:
        if nprocs == 1:
            warnings.warn("Executing only one MPI process! Should be run using" +
                          " e.g. 'mpirun -n X python scriptname.py', where X" +
                          " is the number of parallel processes.")
        if verbose: 
            print("")
            print("number of MPI processes:        ", nprocs)
            print("number of wavelengths:          ", len(sim.efield.wavelengths))
            print("number of wavelengths / process:", int(np.ceil(len(sim.efield.wavelengths) / float(nprocs))))
            print("")
        
        if verbose: print("Generating and splitting jobs... ", end='')
        jobs_all = []
        for i, wl in enumerate(sim.efield.wavelengths):
            ## --- generate simulation objects for each individual wavelength
            _sim = copy.deepcopy(sim)
            _efield = fields.efield(_sim.efield.field_generator, [wl], _sim.efield.kwargs)
            _efield.setDtype(_sim.dtypef, _sim.dtypec)
            _sim.efield = copy.deepcopy(_efield)
            jobs_all.append(_sim)
        jobs = split(jobs_all, nprocs)
        if len(np.unique([len(i) for i in jobs])) > 1:
            warnings.warn("Efficiency warning: Number of wavelengths ({}) ".format(
                                                len(sim.efield.wavelengths)) + 
                          "not divisable by Nr of processes ({})!".format(nprocs))
        if verbose: print("Done.")
    else:
        jobs = None
    
    
    ## --- Scatter jobs across processes and perform GDM simulations for each wavelength
    jobs = comm.scatter(jobs, root=0)
    
    resultsE = []
    resultsH = []
    for job in jobs:
        if verbose: print(" process #{}: Calculating wavelength".format(rank) + \
                               " {}nm".format(job.efield.wavelengths[0]))
        scatter(job, calc_H=calc_H, verbose=scatter_verbose, **kwargs)
        scattered_Efield = job.E
        for _scatf in scattered_Efield:
            resultsE.append(_scatf)
        if calc_H:
            scattered_Hfield = job.H
            for _scatf in scattered_Hfield:
                resultsH.append(_scatf)
    
    
    ## --- Gather results on rank 0
    resultsE = MPI.COMM_WORLD.gather(resultsE, root=0)
    resultsH = MPI.COMM_WORLD.gather(resultsH, root=0)
    
    if comm.rank == 0:
        if verbose: print("All simulations done. Recombining... ", end='')
        ## --- recombine data and sort by wavelength
        resultsE = [i for temp in resultsE for i in temp]
        resultsE = sorted(resultsE, key=lambda k: k[0]['wavelength'])
        sim.E = resultsE
        if calc_H:
            resultsH = [i for temp in resultsH for i in temp]
            resultsH = sorted(resultsH, key=lambda k: k[0]['wavelength'])
            sim.H = resultsH
        if verbose: print("Done.")
    
    return 1
    


# =============================================================================
# Decay rate calculation
# =============================================================================
def _get_K_as_matrix(K, method):
    if method.lower() =='lu':
        warnings.warn("'LU' is not most efficient for explicit matrix inversion. " +
                      "Consider using `scipyinv` instead.")
    ## --- get explicit matrix
    if method.lower() == 'lu':
        import scipy.linalg as la
        K = la.lu_solve(K, np.identity(K[0].shape[0], dtype=K[0].dtype))
    if method.lower() in ['cuda', "cupy"] and type(K) != np.ndarray:
            K = K.get()
    return K


@numba.njit(parallel=True, cache=True)
def _do_double_integral(K, Q, S, chi, Sp):
    for i_probe in numba.prange(Q.shape[0]):
        for i in range(Q.shape[1]):
            for j in range(Q.shape[1]):
                ## matrix multiplication numpy (slow with 'parallel=True')
                # Sp[i_probe] += np.dot(Q[i_probe, i], np.dot(np.dot(chi[i], K[i,j]), S[j, i_probe]))
                
                ## explicit double matrix multiplication (fast with 'parallel=True')
                S_tmp1 = np.zeros((3,3), dtype=np.complex64)
                S_tmp2 = np.zeros((3,3), dtype=np.complex64)
                ## inner matrix multiplication
                for l in range(3): 
                    for m in range(3): 
                        for k in range(3): 
                            S_tmp1[l,m] += chi[i,l,k] * K[i,j,k,m]
                ## middle matrix multiplication
                for l in range(3): 
                    for m in range(3): 
                        for k in range(3): 
                            S_tmp2[l,m] += S_tmp1[l,k] * S[j,i_probe,k,m] 
                ## outer matrix multiplication
                for l in range(3): 
                    for m in range(3): 
                        for k in range(3): 
                            Sp[i_probe,m,l] += Q[i_probe,i,l,k] * S_tmp2[k,m]


def decay_rate(sim, field_index=None, wavelength=None, 
               r_probe=None, r_emitter=None, component='E', 
               return_value='decay_rates', method='scipyinv', verbose=True):
    """local decay rate modification of a electric or magnetic dipole transition
    
    
    Parameters
    ----------
    sim : :class:`.simulation`
        simulation description
    
    field_index : int, default: None
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`.
        Either `field_index` or `wavelength` must be given.
    
    wavelength: float, default: None
        Optional wavelength (alternative to `field_index`) at which to 
        calculate susceptibility matrix (in nm). 
        Either `field_index` or `wavelength` must be given.
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate the decay rate on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
    
    r_emitter : tuple (x,y,z), default: None
        optional fixed coordinate of the emitter, use to evaluate the cross-DOS. 
        If given, the Cross-DOS due to an emitter at `r_emitter` is evaluated 
        at all positions `r_probe`, hence the emitter will be fixed.
        See: *Cazé et al. PRL 110, 063903 (2013)*
    
    component : str, default: 'E'
        "E" or "H": which LDOS component to calculate (electric or magnetic).
    
    return_value : str, default: 'decay_rates'
        Values to be returned. one of:
            - 'decay_rates': relative decayrates for X,Y and Z oriented dipole
            - 'decay_rate_x': relative, partial decayrate for X-oriented dipole
            - 'decay_rate_y': relative, partial decayrate for Y-oriented dipole
            - 'decay_rate_z': relative, partial decayrate for Z-oriented dipole
            - 'decay_rate_total': relative, orientation-averaged decayrate (--> propto Im(Tr(Sp)) )
            - 'field_susceptibility': complex field-susceptibility tensor `Sp` at each r_probe position
    
    method : string, default: "scipyinv"
        inversion method. One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
         - "scipyinv" scipy default inversion (`scipy.linalg.inv`)
         - "numpyinv" numpy inversion (`np.linalg.inv`, if numpy compiled accordingly: LAPACK's `dgesv`)
         - "cupy" uses CUDA GPU via `cupy`
         - "cuda" (equivalent to "cupy")
         - "lu" LU-decomposition (`scipy.linalg.lu_factor`) - inefficient for `decay_rate`!
        
    verbose : bool default=True
        print runtime info
        
    
    Returns:
    ----------
    see parameter `return_value`
    
    
    Notes
    -----
    For details about the underlying formalism, see:
    Wiecha, P. R., Girard, C., Cuche, A., Paillard, V. & Arbouet, A. 
    **Decay Rate of Magnetic Dipoles near Non-magnetic Nanostructures.** 
    Phys. Rev. B 97(8), 085411 (2018)
    
    For details on the concept of the Cross-DOS, see:
    A. Cazé, R. Pierrat & R. Carminati
    **Spatial Coherence in Complex Photonic and Plasmonic Systems**
    Phys. Rev. Lett. 110, 063903 (2013)
    """
# =============================================================================
#     Exception handling
# =============================================================================
    if field_index is None and wavelength is None:
        raise Exception("Either `field_index` or `wavelength` must be given!")
        
    if field_index is not None and wavelength is not None:
        raise Exception("`field_index` AND `wavelength` are given! Ignoring `wavelength`.")
        
    if r_probe is None:
        raise Exception("`r_probe` must be given!.")
        
    if method.lower() not in ["numpyinv", "scipyinv", "lu", "cupy"]:
        raise ValueError('Error: Unknown solving method. Must be one of' +
                         ' ["lu", "numpyinv", "scipyinv", "cuda", "cupy"].')
    
    if r_emitter is not None:
        warnings.warn("Cross-DOS calculation is a new experimental functionality. Use with caution!")
        r_emitter = np.array(r_emitter)
        if np.linalg.norm(sim.struct.geometry - r_emitter).min() <= sim.struct.step:
            raise Exception("`r_emitter` for CDOS must be outside nanostructure. " +
                            "CDOS-emitters inside the structure might get implemented in a future version.")
    
# =============================================================================
# preparation
# =============================================================================
    r_probe = np.array(r_probe)
    if len(r_probe.shape) == 1:
        r_probe = r_probe[None,:]
    if r_probe.shape[0]==3 and r_probe.shape[1]!=3:
        r_probe = np.transpose(r_probe)
    
    if field_index is not None:
        from pyGDM2 import tools
        wavelength = tools.get_field_indices(sim)[field_index]['wavelength']
    
    ## ------- dyads / GDM config
    if component.lower() == 'e':
        ## electric LDOS
        Q = sim.dyads.G_EE
        S = sim.dyads.G_EE
        Q_sign = +1
        calc_H = 0
        ## exclude self-term (assume homogeneous mesh and environment)
        selfterm = sim.dyads.getSelfTermEE(wavelength, sim.struct)[0]
    else:
        ## magnetic LDOS
        Q = sim.dyads.G_HE
        S = sim.dyads.G_HE
        Q_sign = -1
        calc_H = 0
        ## exclude self-term (assume homogeneous mesh and environment)
        selfterm = sim.dyads.getSelfTermHE(wavelength, sim.struct)[0]
    
    ## exclusion distance to avoid divergence of Greens function
    dist_div_G = sim.struct.step
    
    geo = sim.struct.geometry
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
        
    
# =============================================================================
#     the actual calculation
# =============================================================================
    t0 = time.time()
    Sp = np.zeros((len(r_probe), 3, 3), dtype=sim.efield.dtypec)
    
    ## generalized propagator, chi
    if verbose: 
        print("{}-LDOS at wl={:.1f}nm - ".format(component.upper(), wavelength), end='')
    t1 = time.time()
    alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    K = get_general_propagator(sim, wavelength, method=method, calc_H=calc_H)
    K = _get_K_as_matrix(K, method)
    K2 = K.reshape(len(K)//3, 3, -1, 3).swapaxes(1,2).reshape(-1, 3, 3).reshape(len(K)//3,len(K)//3, 3, 3)
    KE2 = K2[:len(geo), :len(geo)]
    # KH2 = K2[len(geo):, :len(geo)]      # K-HE (--> H-field inside)
    if verbose: 
        print("K: {:.1f}s, ".format((time.time() - t1)), end='')
    
# ------------------------------------
# E-LDOS: calc positions in source zone via K
# ------------------------------------
    ## determine positions inside source zone
    if component.lower() != 'e':
        idx_r_outside = np.arange(len(r_probe))
    else:
        from scipy.linalg import norm
        t2 = time.time()
        alpha_inv = np.linalg.inv(alpha_tensor)  # --> list of inverse polarizabilities
        idx_r_outside = []
        for ir, Rp in enumerate(r_probe):
            dist_list = norm(geo - Rp, axis=1)
            idcs_min_dist = np.argsort(dist_list)
            i_in = idcs_min_dist[0]    ## closest meshpoint
            ## --- if inside, use susceptibility at closest meshpoint
            if abs(dist_list[i_in]) <= 1.005*sim.struct.step:
                Sp[ir] = ( KE2[i_in, i_in] - np.eye(3) ) * alpha_inv[i_in]
            else:
                idx_r_outside.append(ir)
        if verbose and len(idx_r_outside)!=len(r_probe): 
            print("source-zone ({}/{} pos): {:.1f}s, ".format(len(r_probe)-len(idx_r_outside), 
                                      len(r_probe), (time.time() - t2)), end='')
    
# ------------------------------------
#     evaluate Green's tensors
# ------------------------------------
    if len(idx_r_outside)>0:
        idx_r_outside = np.array(idx_r_outside)
        _Sp_out = np.zeros((len(idx_r_outside), 3, 3), dtype=sim.efield.dtypec)
        r_probe_out = r_probe.copy()[idx_r_outside]
        
        
        ## light: dipole --> structure
        ## CDOS: fix emitter position, scan only probe position
        if r_emitter is not None:
            r_probe_out = np.ones(r_probe_out.shape) * r_emitter
        t3 = time.time()
        Q_list = np.zeros((len(r_probe_out), len(geo), 3, 3), dtype=sim.efield.dtypec)
        sim.dyads.eval_G(r_probe_out, geo, Q, wavelength, env_dict, Q_list, 
                         selfterm=selfterm, dist_div_G=dist_div_G)
        if verbose: 
            print("Q: {:.1f}s, ".format((time.time() - t3)), end='')
        
        
        ## light: structure --> dipole
        r_probe_out = r_probe.copy()[idx_r_outside]
        t4 = time.time()
        S_list = np.zeros((len(geo), len(r_probe_out), 3, 3), dtype=sim.efield.dtypec)
        sim.dyads.eval_G(geo, r_probe_out, S, wavelength, env_dict, S_list, 
                         selfterm=selfterm, dist_div_G=dist_div_G)
        if verbose: 
            print("S: {:.1f}s, ".format((time.time() - t4)), end='')
        
        
        ## perform double volume integral for each probe position --> numba accelerated
        t5 = time.time()
        _do_double_integral(KE2, Q_sign*Q_list, S_list, alpha_tensor, _Sp_out)
        Sp[idx_r_outside] = _Sp_out
        if verbose: 
            print("integrate: {:.1f}s, ".format((time.time() - t5)), end='')
    if verbose: 
        print("Done in {:.1f}s".format((time.time() - t0)))
    
    
# ------------------------------------
#     return results
# ------------------------------------
    if return_value.lower() == 'field_susceptibility':
        return Sp
    else:
        # =====================================================================
        # evaluate field-susceptibilities
        # =====================================================================
        k0 = 2*np.pi/wavelength
        gamma_0 = 1
        if return_value.lower() == 'decay_rates':
            dp_list = np.array([[1,0,0], [0,1,0], [0,0,1]])
        if return_value.lower() == 'decay_rate_x':
            dp_list = np.array([[1,0,0]])
        if return_value.lower() == 'decay_rate_y':
            dp_list = np.array([[0,1,0]])
        if return_value.lower() == 'decay_rate_z':
            dp_list = np.array([[0,0,1]])
        
        if return_value.lower() not in ['decay_rate_tot', 'decay_rate_total']:
            gamma_scalar_maps = []
            for i_mu, mu in enumerate(dp_list):
                l_mu = np.linalg.norm(mu)
                mu_e = mu / l_mu
                
                gamma = np.zeros(len(r_probe), dtype=sim.dtypef)
                for i, Sp_i in enumerate(Sp):
                    gamma[i] = (gamma_0 + (3./2.) * (1./k0**3) * gamma_0 * 
                                    l_mu * np.dot(np.dot(mu_e, Sp_i.imag), mu_e))
            
                gamma_scalar_maps.append(np.concatenate([r_probe, gamma[:,None]], axis=1))
            return gamma_scalar_maps
        else:
            Sp_tot = (gamma_0 + (3./2.) * (1./k0**3) * gamma_0 * np.trace(Sp.imag, axis1=1, axis2=2))
            return np.concatenate([r_probe, Sp_tot[:,None]], axis=1)





#==============================================================================
# Matrix operations
#==============================================================================
def _get_gpu_memory_subprocess(which_gpu=0):
    """
    from:
    https://stackoverflow.com/questions/59567226/how-to-programmatically-determine-available-gpu-memory-with-tensorflow
    """
    import subprocess as sp
    _output_to_list = lambda x: x.decode('ascii').split('\n')[:-1]

    COMMAND = "nvidia-smi --query-gpu=memory.free --format=csv"
    memory_free_info = _output_to_list(sp.check_output(COMMAND.split()))[1:]
    memory_free_values = [int(x.split()[which_gpu]) for i, x in enumerate(memory_free_info)]

    return memory_free_values[0]


def _get_gpu_memory_nvidia_module(which_gpu=0):
    """
    from:
    https://stackoverflow.com/questions/59567226/how-to-programmatically-determine-available-gpu-memory-with-tensorflow
    """
    import nvidia_smi
    
    nvidia_smi.nvmlInit()
    handle = nvidia_smi.nvmlDeviceGetHandleByIndex(which_gpu)
    info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)
    
    return info.free / 1024**2  # in MiB


def _get_free_GPU_memory():
    """returns free GPU memory in MiB (nvidia only)"""
    try:
        free_mem = _get_gpu_memory_nvidia_module()
    except: 
        try:
            free_mem = _get_gpu_memory_subprocess()
        except:
            return -1
    return free_mem
            


def get_general_propagator(sim=None, wavelength=None, method='lu', 
                           fallbackmethod_gpu='lu',
                           calc_H=False, verbose=False):
    """invert dipole-coupling matrix
    
    Parameters
    ----------
    sim : :class:`.simulation`
        simulation description
    
    wavelength: float
        Wavelength at which to calculate susceptibility matrix (in nm)
    
    method : string, default: "lu"
        inversion method. One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
         - "lu" LU-decomposition (`scipy.linalg.lu_factor`)
         - "scipyinv" scipy default inversion (`scipy.linalg.inv`)
         - "numpyinv" numpy inversion (`np.linalg.inv`, if numpy compiled accordingly: LAPACK's `dgesv`)
         - "cupy" uses CUDA GPU via `cupy`
         - "cuda" (equivalent to "cupy")
    
    fallbackmethod_gpu : string, default: "lu"
        method to fall back to if using cuda solver and GPU runs out of RAM 
    
    calc_H : bool, default: False
        if True, calculate also the internal magnetic field H. This has an 
        impact on performance and memory requirement, so it is deactivated by 
        default.
    
    verbose : bool, default False
        Print timing info to stdout
        
    
    Returns
    -------
      - K: Generalized Propagator
    
    
    Notes
    -----
    For details on the concept of the generalized propagator, see e.g.:
    Martin, O. J. F. & Girard, C. & Dereux, A. **Generalized Field Propagator 
    for Electromagnetic Scattering and Light Confinement.**
    Phys. Rev. Lett. 74, 526–529 (1995).
    
    For the Electric-magnetic mixed field calculation, see e.g.:
    Schröter, U. **Modelling of magnetic effects in near-field optics.** 
    Eur. Phys. J. B 33, 297–310 (2003).
    """
# =============================================================================
#     Exception handling
# =============================================================================
    if method.lower() == 'lu' and calc_H:
        warnings.warn("calculating 'H' is not efficient with LU decomposition. " +
                      "Consider using another method like `scipyinv` instead.")
    
    if method.lower() == "cuda":
        method = "cupy"
        
    if method.lower() == "cupy":
        import cupy
        if int(cupy.__version__.split('.')[0])<7:
            raise ValueError("`cupy` version 7 or higher required. Found cupy " +
                             "version {}. Please upgrade cupy.".format(cupy.__version__))
    
    if method.lower() not in ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]:
        raise ValueError('Invalid inversion method. Must be one of ["lu", ' +
                         '"numpyinv", "scipyinv", "cupy", "cuda"].')
    
    
# =============================================================================
#     setup EE coupling matrix
# =============================================================================
    t0 = time.time()
    ## --- construct matrix
    M_EE = get_SBS_EE(sim, wavelength=wavelength, invertible=True)                
    
    if verbose: 
        print('timing for wl={:.2f}nm - setup: EE {:.1f}ms, '.format(
                            wavelength, 1000.*(time.time()-t0)), end='')
    t0b = time.time()
    
# =============================================================================
#    inversion of EE part (via chosen method)
# =============================================================================
    ## --- CUDA based inversion on GPU via `cupy`
    if method.lower() == "cupy":
        import cupy as cp
        import cupy.linalg
        ## free no longer used GPU RAM
        mempool = cp.get_default_memory_pool()
        mempool.free_all_blocks()
        
        ## check available memory:
        free_mem = _get_free_GPU_memory()
        if free_mem == -1:
            warnings.warn("Failed checking available GPU memory. Install `nvidia-ml-py3` (e.g. via pip) if you want the GPU-RAM test to work.")
        else:
            req_mem = 4 * M_EE.nbytes/1024**2    # 'inv' requires memorey of 4*array
            if req_mem > free_mem:
                warnings.warn("Required memory exceeds GPU RAM. Falling back to scipy's 'scipyinv'. You could also try 'LU' instead, which is often more efficient.")
                method = fallbackmethod_gpu
            else:
                ## move array to cuda device, cuda inversion
                Agpu = cp.array(np.ascontiguousarray(M_EE))
                Ainv_gpu = cupy.linalg.inv(Agpu)
                K_EE = Ainv_gpu
                
                mempool.free_all_blocks()    # free no longer used GPU RAM
    
    
    ## --- CPU inversion methods
    if method.lower() in ["numpyinv", "scipyinv", "lu"]:
        if method.lower() in ["numpyinv"]:     # pyGDM2 default
            K_EE = np.linalg.inv(M_EE)
        elif method.lower() in ["scipyinv"]:
            import scipy.linalg as la
            K_EE = la.inv(M_EE, overwrite_a=True)
        elif method.lower() in ["lu"]:
            import scipy.linalg as la
            K_EE = la.lu_factor(M_EE, overwrite_a=True)
        del M_EE#; gc.collect()
    
    
# =============================================================================
#     get HE part of propagator and calc full K via block-inversion
# =============================================================================
    t1 = time.time()
    if calc_H:
        M_HE = get_SBS_HE(sim, wavelength)
        
        if method.lower() == 'lu':
            import scipy.linalg as la
            K_EE = la.lu_solve(K_EE, np.identity(3*len(sim.struct.geometry), 
                                                 dtype=K_EE[0].dtype))
        
        if method.lower() == 'cupy':
            import cupy as cp
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()    # free no longer used GPU RAM
            K_HE = -1*cp.dot(cp.array(np.ascontiguousarray(M_HE)), K_EE)
            mempool.free_all_blocks()    # free no longer used GPU RAM       
            K_HE = K_HE.get()
            K_EE = K_EE.get()
        else:
            K_HE = -1.*np.dot(M_HE, K_EE)
        
        M_zero     = np.zeros(2*[3*len(sim.struct.geometry)], dtype=sim.dtypec)
        M_I        = np.identity(3*len(sim.struct.geometry), dtype=sim.dtypec)
        
        ## -- block inversion K = [[M_EE^-1, 0], [-1*M_HE*M_EE^-1, Identity]]
        K = np.block([[K_EE, M_zero],
                      [K_HE, M_I]])
        if verbose: 
            print('HE {:.1f}ms, '.format(1000.*(time.time()-t1)), end='')
    else:
        K = K_EE
    
    ## -- done
    if verbose: 
        print('inv.: {:.1f}ms, '.format(1000.*(t1-t0b)), end='')
    
    return K




# =============================================================================
# CPU - side-by-side matrix setup
# =============================================================================
def get_SBS_EE(sim, wavelength, invertible=True):
    ##  --- simulation config
    geo = sim.struct.geometry
    
    ## --- material permittivity related config
    self_term_tensors = sim.dyads.getSelfTermEE(wavelength, sim.struct)
    alpha_tensors = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)

    ## --- run numba routine for EE coupling
    M_EE = np.zeros((len(geo)*3,len(geo)*3), dtype=sim.dtypec)
    conf_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    sim.dyads.tsbs_EE(geo, wavelength, self_term_tensors, alpha_tensors, 
                      conf_dict, M_EE)
    
    N = np.shape(M_EE)[0]
    if not invertible:
        M_EE = np.asfortranarray(np.identity(N) - M_EE, dtype=sim.dtypec)
    
    return M_EE


def get_SBS_HE(sim, wavelength):
    ##  --- simulation config
    geo = sim.struct.geometry
    
    ## --- material permittivity related config
    self_term_tensors = sim.dyads.getSelfTermHE(wavelength, sim.struct)
    alpha_tensors = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    
    ## --- run numba routine for EH coupling
    M_HE = np.zeros((len(geo)*3,len(geo)*3), dtype=sim.dtypec)
    conf_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    sim.dyads.tsbs_HE(geo, wavelength, self_term_tensors, alpha_tensors, 
                      conf_dict, M_HE)
    
    return M_HE










#==============================================================================
# Internal Helper Functions
#==============================================================================
def _superVectorToFieldList(E):
    """convert complex 3N supervector E to list of N field tuples (Ex,Ey,Ez)"""
    return np.reshape(E, (int(len(E)/3), 3))


def _fieldListToSuperVector(E):
    """convert list of N field tuples (Ex,Ey,Ez) to complex 3N supervector E"""
    return np.reshape(E, (np.product(E.shape)))









#%%
if __name__ == "__main__":
    pass
    




# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Load the result of the "own problem" optimization, calculate the 
    scattering spectrum and plot to verify that a resonant structure was found.

"""


import numpy as np
import matplotlib.pyplot as plt
import copy


from pyGDM2 import linear
from pyGDM2 import tools
from pyGDM2 import visu
from pyGDM2 import structures
from pyGDM2 import materials

from pyGDM2 import core
from pyGDM2 import fields

from pyGDM2.EO.tools import get_pareto_fronts, plot_pareto_2d





#==============================================================================
# Load best candidate from optimization
#==============================================================================
## --- optimization file locations
results_folder = 'eo_out'
results_suffix = 'test_MO_problem'



## --- "get_pareto_fronts" needs the problem definition in namespace
from example_eo_07_multi_objective_problem import ownMultiObjectiveProblem

paretos, sims = get_pareto_fronts(results_folder, results_suffix, iteration=-1, verbose=True)

pareto_f = paretos[0]
sims_f   = sims[0]
testidx = 9

sim = sims_f[testidx]
plot_pareto_2d(pareto_f)

print "structure #", testidx
visu.structure(sim)

#%%

#==============================================================================
# setup new simulation to calculate spectrum
#==============================================================================
## --- structure
struct = copy.deepcopy(sim.struct)

## --- incident field
field_generator = fields.planewave        # planwave excitation
wavelengths = np.arange(400, 1615, 30)  # spectrum
kwargs = dict(theta = [0.0, 90.0])          # 0/90 deg polarizations
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## --- simulation
sim_spectrum = core.simulation(struct, efield)


#==============================================================================
# run simulation for the spectrum
#==============================================================================
core.scatter(sim_spectrum, verbose=1)
#%%

## spectrum
field_kwargs = tools.get_possible_field_params_spectra(sim_spectrum)[0]
wl, spec_ext0 = tools.calculate_spectrum(sim_spectrum, field_kwargs, linear.extinct)
asca0 = spec_ext0.T[1]

field_kwargs = tools.get_possible_field_params_spectra(sim_spectrum)[1]
wl, spec_ext90 = tools.calculate_spectrum(sim_spectrum, field_kwargs, linear.extinct)
asca90 = spec_ext90.T[1]

geom_cs = tools.get_geometric_cross_section(sim_spectrum)



#%%
## --- plot
plt.figure(figsize=(10,5))

## --- spectra
plt.subplot2grid((1,5), (0,0), colspan=3)
plt.title("scattering spectrum")
plt.plot(wl, asca0/geom_cs, label="0deg")
plt.plot(wl, asca90/geom_cs, label="90deg")

plt.legend(loc='best', fontsize=10)
plt.xlabel("wavelength (nm)")
plt.ylabel("Q_scat")


## --- structure
plt.subplot2grid((1,5), (0,3), colspan=2, aspect="equal")
plt.title('structure geometry')
visu.structure(sim_spectrum, show=False)
    
plt.show()





#%%
tools.save_simulation(sim_spectrum, "simulation_mo_gold_cross_f1_{:.2f}_f2_{:.2f}_spec.pcl".format( 
                            pareto_f[testidx][0], pareto_f[testidx][1] ))

print np.round(pareto_f[testidx],2)











# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
fast electron fields

authors: A. Arbouet, P. Wiecha
"""

from __future__ import print_function
from __future__ import absolute_import

import warnings

import numpy as np



#==============================================================================
# globals
#==============================================================================
DTYPE_C = np.complex64




        

#==============================================================================
# field generator functions
#==============================================================================
def fast_electron(pos, env_dict, wavelength, 
                  electron_kinetic_energy, x0, y0, 
                  kSign = -1, avoid_div_lim_distance=10.0,
                  returnField='E'):
    
    """Electric field created by a fast electron moving along (OZ) 
    
    The electron beam crosses the (OXY) plane in (x0, y0)
       
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain either "eps_env" or ["eps1", "eps2"].
    
    wavelength : float
        Wavelength in nm
    
    electron_kinetic_energy : float
        electron kinetic energy (keV)
    
    x0, y0 : float
        position of the electron beam (nm)
    
    kSign : int, default: -1
        sign of wavenumber. 
        +1: electron propagation from bottom to top (towards increasing z)
        -1: electron propagation from top to bottom (towards smaller z, default)
        either kSign or k0 must be given.
    
    avoid_div_lim_distance : float, default: 10.0
        set a min. distance (in nm) between the location where the E-field 
        is computed and the electron trajectory to avoid divergence  
    
    returnField : str, default: 'E'
        if 'E': returns electric field; if 'B': magnetic field
    
    Returns
    -------
      E0 (B0):       Complex E-(B-)Field at each dipole position as 
                     list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]
    """
    from scipy import special
    
    if 'eps_env' in env_dict.keys():
        n2 = n2 = n3 = env_dict['eps_env']**0.5
    else:
        # n1 = env_dict['eps1']**0.5
        n2 = env_dict['eps2']**0.5
        n3 = env_dict['eps3']**0.5
        # spacing = np.float32(env_dict['spacing'].real)
        if n2 != n3:
            warnings.warn("fast_electron only supports a single interface " + 
                          "(between `n1`/`n2`). " +
                          "The simulation might not be a good approximation.")
    
    ## constant parameters
    qe = 4.80321E-10     # elementary charge (Franklin)    
    Eo_el = 511.0        # electron rest mass (keV)	   
    c = 2.99792458E10    # Speed of light (cm/s)   

    gamma = 1. + electron_kinetic_energy/Eo_el  # Gamma = Lorentz factor
    f = np.sqrt(1.-1./gamma**2.)                # speed of electrons in units of c
    kz =  n2*(2.*np.pi/wavelength)            
    

    Ex = np.zeros(len(pos), dtype=DTYPE_C)
    Ey = np.zeros(len(pos), dtype=DTYPE_C)
    Ez = np.zeros(len(pos), dtype=DTYPE_C)
    for ipos, rp in enumerate(pos):
        xpos = rp[0] - x0
        ypos = rp[1] - y0   
        zpos = rp[2]    
        ##  --------- Electric field --------- 
        R = np.sqrt(xpos**2 + ypos**2)
        if (R > avoid_div_lim_distance):
            if returnField.lower() == 'e':
                U = kz*R / (f*gamma)
                phase = np.exp(1j*kSign*kz*zpos/f)
                factor = qe * kz * 1E7 / (np.pi * c * f**2. * gamma * n2**2.)
                Er = phase * factor * special.kv(1, U)
                Ex[ipos] = - Er * xpos/R
                Ey[ipos] = - Er * ypos/R
                Ez[ipos] = kSign*factor*1j * special.kv(0, U)/gamma
        ##  --------- Magnetic field --------- 
            else:
                ## !!! TODO
                raise NotImplementedError("fast-electron magnetic field not yet implemented.")
                
        else:
            Ex[ipos] = 0.0
            Ey[ipos] = 0.0
            Ez[ipos] = 0.0
    
    return np.transpose([Ex, Ey, Ez])




if __name__ == "__main__":
    pass



# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Demonstrate how to implement a user-defined problem. 
    The example problem is the maximization of the scattering cross-section
    of a simple plasmonic nano-structure under plane wave illumination with 
    fixed wavelength and linear polarization angle.

"""

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core

from pyGDM2.EO.problems import ProblemMaxScat
from pyGDM2.EO.core import do_eo


from PyGMO import algorithm



#==============================================================================
# Implement a simple problem: Maximize scattering cross-section
#==============================================================================
from pyGDM2.EO.models import BaseModel

class ownModel(BaseModel):
    """optimization-model for simple geometry, consisting of one rectangular antenna
    
    Note : 
        The purpose of this model is only for demonstration. A problem with only 
        two free parameters should probably be solved by brute-force (simulation of 
        all possible permutations)
    
    """
    def __init__(self, sim, limits_W, limits_L, height):
        """
        """
        ## --- init basemodel with simulation instance
        super(self.__class__, self).__init__(sim)
        
        ## --- width and length limits for rectrangular antenna
        self.limits_W = limits_W
        self.limits_L = limits_L
        
        ## --- fixed parameters
        self.height = height    # height of rectangular antenna
        
        ## --- init with random values. `set_step`and `random_structure` are 
        ##     defined in `BaseModel`
        self.set_step(self.sim.struct.step)
        self.random_structure()
    
    def get_dim(self):
        """two free parameters: width / length"""
        return 2
    
    def get_bounds(self):
        """Return lower and upper boundaries for parameters as required by pyGMO models"""
        self.lbnds = [self.limits_W[0], self.limits_L[0]]
        self.ubnds = [self.limits_W[1], self.limits_L[1]]
        
        return self.lbnds, self.ubnds
    
    def generate_structure(self, params):
        """generate the structure"""
        ## --- order of `params` must correspond to boundary order defined by `get_bounds`
        W, L = params[0], params[1]
        
        from pyGDM2 import structures
        geometry = structures.rectwire(self.sim.struct.step, L, self.height, W)
        
        ## -- set new structure-geometry (`set_structure` is a `BaseModel` function)
        self.set_structure(geometry)

















if __name__ == '__main__':
    #==============================================================================
    # Setup pyGDM part
    #==============================================================================
    ## ---------- Setup structure
    mesh = 'cube'
    step = 15
    material = materials.dummy(2.0)    # material: constant and real dielectric function
    material = materials.gold()        # material: gold
    n1, n2 = 1.0, 1.0         # constant environment
    
    ## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
    geometry = []       
    
    struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))
    
    
    
    ## ---------- Setup incident field
    field_generator = fields.planewave        # planwave excitation
    kwargs = dict(theta = [0.0])              # several polarizations
    wavelengths = [800]                       # one single wavelength
    efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)
    
    
    
    ## ---------- Simulation initialization
    sim = core.simulation(struct, efield)
    
    
    
    
    #==============================================================================
    # setup evolutionary optimization
    #==============================================================================
    ## --- output folder and file-names
    results_folder = 'eo_out'
    results_suffix = 'test_own_model_Q_90'
    target = 'Qscat'
    results_suffix = 'test_own_model_Q_800'
    target = 'Qscat'
    
#    results_suffix = 'test_own_model_CS'
#    target = 'CSscat'
    
#    results_suffix = 'test_own_model_CS_800'
#    target = 'CSscat'
    
    
    ## --- structure model and optimizaiton problem
    limits_W = [2, 20]
    limits_L = [2, 20] 
    height = 3
    model = ownModel(sim, limits_W, limits_L, height)
    
    problem_dimension = model.get_dim()
    problem = ProblemMaxScat(problem_dimension, model, opt_target=target)
    
    
    ## --- algorithm configuration
    algo = algorithm.jde(gen=1, memory=True)
    
    islands = 1             # number of islands ( > 1: multi-processing)
    population = 20         # individuals per island
    
    
    ## --- stop criteria
    max_time = 1800         # seconds
    max_iter = 30          # max. iterations
    max_nonsuccess = 10     # max. consecutive iterations without improvement
    
    
    
    
    
    #==============================================================================
    # run evolution
    #==============================================================================
    #do_eo(results_folder, results_suffix, problem)    # run with default params
    do_eo(results_folder, results_suffix,
          problem, algo, 
          islands, population,
          max_time, max_iter, max_nonsuccess, 
          ## --- flags
          runtime_plotting=True,
          save_each_generation=False,
          verbose=True)
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    



from __future__ import print_function, division

import cmath

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import fields
from pyGDM2 import tools
from pyGDM2 import materials
from pyGDM2 import visu

#%%
## incident field generato --> code 3-layer of Christian
field_generator = fields.evanescent_planewave

## wavelegnths / angles to evaluate
wavelengths = np.linspace(400, 800, 41)
teta_list = np.linspace(35,55, 35)

## layer material and thickness
layer_material = materials.gold()
#layer_material = materials.silicon()
spacing = 50      # layer thinkness

## environment
n1 = 1.5  # substrate
n3 = 1.0  # top


## 2D evaluation plane
projection = 'XZ'

## transmitted --> integrate intensity in some region far above layer
r_probe_T = tools.generate_NF_map(-1000,1000,11, 3500,4500,11,0, projection=projection)

## reflected --> integrate intensity in some region below layer
r_probe_R = tools.generate_NF_map(-1000,1000,11, -4500,-3500,11,0, projection=projection)



# =============================================================================
# evaluate fields
# =============================================================================
I_T = []
I_R = []
## loop over wavelengths
for wl in wavelengths:
    _IT = []
    _IR = []
    ## loop over incident angles
    for t in teta_list:
        kwargs = dict(theta_inc=t, polar='p')

        ## gold refractive index at current wavelength
        n2 = cmath.sqrt(layer_material.epsilon(wl))
        
        ## NF above (transmission)
        NF_T = tools.evaluate_incident_field(field_generator, wl, kwargs, r_probe_T,
                                           n1=n1,n2=n2,n3=n3, spacing=spacing)
        
        ## NF below (incidence + reflection)
        NF_below = tools.evaluate_incident_field(field_generator, wl, kwargs, r_probe_R,
                                           n1=n1,n2=n2,n3=n3, spacing=spacing)
        ## NF_inc0: get incident field in a homogeneous environment (substrate ref.index)
        NF_inc0 = tools.evaluate_incident_field(field_generator, wl, kwargs, r_probe_R,
                                           n1=n1,n2=n1,n3=n1, spacing=spacing)
        ## reflected = NF_inc - NF_inc0
        NF_R = NF_below.copy()
        NF_R[:,3:] -= NF_inc0[:,3:]
        
        ## calculate intensity at each point and take average of entire evaluation region
        I_tot_T = np.average(np.sum( np.abs(NF_T[:,3:])**2, axis=1 ))
        I_tot_R = np.average(np.sum( np.abs(NF_R[:,3:])**2, axis=1 ))
        _IT.append(I_tot_T)
        _IR.append(I_tot_R)
    I_T.append(_IT)
    I_R.append(_IR)
I_T = np.flipud(I_T)
I_R = np.flipud(I_R)
    
#%%
# =============================================================================
# plot results
# =============================================================================

plt.figure(figsize=(12,3))
## --- field intensity transmitted
plt.subplot(131)
plt.title("transmitted")
plt.imshow(I_T, extent=(min(teta_list), max(teta_list), min(wavelengths), max(wavelengths)), aspect='auto')
plt.colorbar(label="|E|^2 / |E0|^2")
plt.xlabel("incident angle (deg)")
plt.ylabel("wavelength (nm)")

## --- field intensity reflected
plt.subplot(132)
plt.title("reflected")
plt.imshow(I_R, extent=(min(teta_list), max(teta_list), min(wavelengths), max(wavelengths)), aspect='auto')
plt.colorbar(label="|E|^2 / |E0|^2 above")
plt.xlabel("incident angle (deg)")
plt.ylabel("wavelength (nm)")

## --- one selected wavelength
plt.subplot(133)
plt.title("wavelength = {}nm".format(wavelengths[40]))
plt.plot(teta_list, I_R[40])
plt.xlabel("incident angle (deg)")
plt.ylabel("reflectance")


plt.tight_layout()
#plt.savefig("surface_plasmon_gold.png")
plt.show()

#%%
# =============================================================================
# plot real part of electric field across layer
# =============================================================================
wl = 600
teta = 45
kwargs = dict(theta_inc=teta, polar='p')
n2 = cmath.sqrt(layer_material.epsilon(wl))


#r_probe = tools.generate_NF_map(-1000,1000,100, -200,350,50,0, p7rojection=projection)
r_probe = tools.generate_NF_map(-2000,2000,61, -1500,1500,41,0, projection=projection)

NF = tools.evaluate_incident_field(field_generator, wl, kwargs, r_probe,
                                           n1=n1,n2=n2,n3=n3, spacing=spacing)

v = visu.vectorfield(NF, complex_part='real', projection=projection,
                     tit=projection+' real part, full (inc+r)', show=1, scale=15)




#%%
# =============================================================================
# 3D animation
# =============================================================================
#from pyGDM2 import visu3d
#visu3d.animate_vectorfield(NF, scale=10)



# encoding: utf-8
"""
Created on June 1, 2017

@author: P. R. Wiecha

"""
from __future__ import print_function, division

from pyGDM2 import core
from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import linear
from pyGDM2 import tools
from pyGDM2 import visu

from pyGDM2.EO.core import run_eo
from pyGDM2.EO.problems import ProblemScat
from pyGDM2.EO.models import RectangularAntenna
from pyGDM2.EO.tools import get_best_candidate

import matplotlib.pyplot as plt
import numpy as np
import copy



#==============================================================================
# Setup pyGDM part
#==============================================================================
## ---------- Setup structure
mesh = 'cube'
step = 15
material = materials.gold()   # structure material
n1, n2 = 1.0, 1.0             # vacuum environment

## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
geometry = []       
struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))


## ---------- Setup incident field
field_generator = fields.planewave        # planwave excitation
kwargs = dict(theta = [0.0])              # target lin. polarization angle
wavelengths = [1000]                      # target wavelength
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## ---------- Simulation initialization
sim = core.simulation(struct, efield)


#==============================================================================
# setup evolutionary optimization
#==============================================================================
## --- structure model: Rectangular planar antenna of fixed height
limits_W   = [2, 20]  # units of "step"
limits_L   = [2, 20]  # units of "step"
limits_pos = []   # units of nm, [] -->  no shift allowed
height = 3    # units of "step"
model = RectangularAntenna(sim, limits_W, limits_L, limits_pos, height)

## --- optimization problem: Scattering
opt_target = 'Qscat'  # 'Qscat' --> scat. efficiency
problem = ProblemScat(model, opt_target=opt_target)


## --- filename to save results 
results_filename = 'eo_Qscat.eo'

## --- size of population
population = 15          # Nr of individuals

## --- stop criteria
max_time = 60            # seconds
max_iter = 20            # max. iterations
max_nonsuccess = 5      # max. consecutive iterations without improvement

## --- other config
generations = 1          # generations to evolve between status reports
plot_interval = 5        # plot each N improvements
save_all_generations = False

##  Use algorithm "sade" (jDE variant, a self-adaptive form of differential evolution)
import pygmo as pg
algorithm = pg.sade
algorithm_kwargs = dict()   # optional kwargs passed to the algorithm


eo_dict = run_eo(problem,
                 population=population,
                 algorithm=algorithm,
                 plot_interval=plot_interval, 
                 generations=generations, 
                 max_time=max_time, max_iter=max_iter, max_nonsuccess=max_nonsuccess,
                 filename=results_filename)



# =============================================================================
# analyze and show results
# =============================================================================
## get simulation of best candidate
sim = get_best_candidate(eo_dict, iteration=-1, verbose=True)


#==============================================================================
# setup new simulation to calculate spectrum
#==============================================================================
## --- structure
struct = copy.deepcopy(sim.struct)


## --- incident field
field_generator = fields.planewave        # planwave excitation
wavelengths = np.arange(600, 1410, 20)    # spectrum
kwargs = dict(theta = [0.0, 90.0])        # 0/90 deg polarizations
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## --- simulation
sim_spectrum = core.simulation(struct, efield)


#==============================================================================
# run simulation for the spectrum
#==============================================================================
core.scatter(sim_spectrum, method='lu')


## spectrum
wl, spec_ext0 = tools.calculate_spectrum(sim_spectrum, 0, linear.extinct)
asca0 = spec_ext0.T[1]

wl, spec_ext90 = tools.calculate_spectrum(sim_spectrum, 1, linear.extinct)
asca90 = spec_ext90.T[1]

geom_cs = tools.get_geometric_cross_section(sim_spectrum)


#%%
## --- plot
plt.figure(figsize=(10,5))

## --- spectra
plt.subplot2grid((1,5), (0,0), colspan=3)
plt.title("scattering spectrum")
plt.plot(wl, asca0/geom_cs, label="0deg")
plt.plot(wl, asca90/geom_cs, label="90deg")

plt.legend(loc='best', fontsize=10)
plt.xlabel("wavelength (nm)")
plt.ylabel("Q_scat")


## --- structure
plt.subplot2grid((1,5), (0,3), colspan=2, aspect="equal")
plt.title('structure geometry')
visu.structure(sim_spectrum, show=False)
    
plt.show()







